# Generated by Django 3.0.6 on 2020-05-09 22:18

import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0002_histograms'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProviderNotification',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('key', models.CharField(help_text='Notification unique key for the provider', max_length=250)),
                ('info', models.TextField(help_text='Notification information.', null=True)),
                ('when', models.DateTimeField(help_text='When the notification should be sent to provider.')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('provider', models.ForeignKey(help_text='Related provider', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Provider')),
            ],
            options={
                'unique_together': {('provider', 'key')},
            },
        ),
    ]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Course Group views module """
from django.shortcuts import get_object_or_404

from django_filters import rest_framework as filters_dj

from django_filters.rest_framework import DjangoFilterBackend

from rest_framework import viewsets
from rest_framework.filters import OrderingFilter
from rest_framework_extensions.mixins import NestedViewSetMixin

from tesla_ce.apps.api.v2.serializers import InstitutionCourseGroupSerializer
from tesla_ce.apps.api.v2.serializers import InstitutionCourseGroupCourseSerializer
from tesla_ce.models import Course
from tesla_ce.models import CourseGroup

def groups_values():
    groups = CourseGroup.objects.all()
    return ((group.id, group) for group in groups)


class GroupFilter(filters_dj.FilterSet):
    parent = filters_dj.ChoiceFilter(field_name="parent",
                                     label="parent",
                                     choices=groups_values,
                                     null_label='None',
                                     null_value=-1,
                                     method='filter_parent')

    name = filters_dj.CharFilter(lookup_expr='icontains')
    description = filters_dj.CharFilter(lookup_expr='icontains')

    def filter_parent(self, queryset, name, value):

        # In case value is not a valid integer, ignore the filter
        try:
            value = int(value)
        except ValueError:
            return queryset

        # Request users with no institution assigned
        if value < 0:
            return queryset.filter(parent=None)

        # Filter by provided institution
        return queryset.filter(parent_id=value)

    class Meta:
        model = CourseGroup
        fields = ['name', 'description', 'parent']


# pylint: disable=too-many-ancestors
class InstitutionCourseGroupViewSet(NestedViewSetMixin, viewsets.ModelViewSet):
    """
    API endpoint that allows course groups to be viewed or edited.
    """
    model = CourseGroup
    serializer_class = InstitutionCourseGroupSerializer
    filter_backends = [DjangoFilterBackend, OrderingFilter]
    filterset_fields = ['parent', 'name', 'description']
    filterset_class = GroupFilter

    def get_queryset(self):
        queryset = super().filter_queryset_by_parents_lookups(CourseGroup.objects)
        return queryset.all().order_by('id')


# pylint: disable=too-many-ancestors
class InstitutionCourseGroupCourseViewSet(NestedViewSetMixin, viewsets.ModelViewSet):
    """
    API endpoint that allows courses in course groups to be added or deleted.
    """
    model = Course
    serializer_class = InstitutionCourseGroupCourseSerializer
    filter_backends = [DjangoFilterBackend, OrderingFilter]
    filterset_fields = ['vle_id', 'vle_course_id', 'description', 'code', 'start', 'end']

    def get_queryset(self):
        """
            Get the queryset for this view
            :return: Queryset
        """
        group = get_object_or_404(super().filter_queryset_by_parents_lookups(CourseGroup.objects))
        return group.courses.all().order_by('id')

    def perform_destroy(self, instance):
        """
            Remove the course from the group
            :param instance: Course
        """
        group = get_object_or_404(super().filter_queryset_by_parents_lookups(CourseGroup.objects))
        group.courses.remove(instance)



# Generated by Django 3.2.4 on 2021-07-16 08:05

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0017_auto_20210715_2150'),
    ]

    operations = [
        migrations.AddField(
            model_name='assessmentsession',
            name='auto_closed',
            field=models.BooleanField(default=None, help_text='Whether this session has been automatically closed', null=True),
        ),
    ]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Learner profile views module """
import json

from rest_framework.views import APIView
from rest_framework.views import Response
from rest_framework.views import status

from tesla_ce.models import AuthenticatedUser
from tesla_ce.models import Learner
from tesla_ce.models import InstitutionUser

from tesla_ce.lib.exception import tesla_report_exception

from ..utils import is_authenticated
from ..serializers import ProfileSerializer


class ProfileView(APIView):
    """
        Base class for Learner API views
    """
    serializer_class = ProfileSerializer

    def get(self, request, institution_id, learner_id):
        """
            Get learner profile
            :param request: The received request
            :param institution_id: Institution unique ID
            :param learner_id: Learner ID in TeSLA
        """
        try:
            # Check that provided token is compatible with the data
            if not is_authenticated(request, institution_id, learner_id, {'learner_id': str(learner_id)}):
                return Response(status=status.HTTP_403_FORBIDDEN, data={'status': "ERROR", 'errors': "Invalid token"})
            if isinstance(request.user, AuthenticatedUser):
                return Response(status=status.HTTP_403_FORBIDDEN,
                                data={'status': "OK", 'profile': {
                                   'learner_id': str(learner_id)
                                }})
            if isinstance(request.user, Learner):
                return Response(ProfileSerializer(request.user).data)
            if isinstance(request.user, InstitutionUser):
                return Response(ProfileSerializer(request.user.learner).data)

            return Response(status=status.HTTP_403_FORBIDDEN, data={'status': "ERROR", 'errors': "Invalid user object"})

        except Exception as ex:
            tesla_report_exception(ex)
            return Response(status=status.HTTP_500_INTERNAL_SERVER_ERROR,
                            data={"status": "ERROR", 'errors': ex},
                            content_type="application/json")



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Provider notification serializers module """
from rest_framework import serializers

from tesla_ce.apps.api.utils import JSONField
from tesla_ce.models import ProviderNotification


class ProviderNotificationSerializer(serializers.ModelSerializer):
    """Provider notification serialize class."""
    info = JSONField(allow_null=True, default=None)
    provider_id = serializers.HiddenField(default=None)
    provider = serializers.IntegerField(read_only=True, source="provider_id")

    class Meta:
        model = ProviderNotification
        fields = "__all__"

    def validate(self, attrs):
        """
            Validate the given attributes
            :param attrs: Attributes parsed from request
            :type attrs: dict
            :return: Validated attributes
            :rtype: dict
        """
        # Add predefined fields
        attrs['provider_id'] = self.context['view'].kwargs['parent_lookup_provider_id']

        try:
            self.instance = ProviderNotification.objects.get(provider_id=attrs['provider_id'],
                                                             key=attrs['key'])
        except ProviderNotification.DoesNotExist:
            pass

        return super().validate(attrs)

    def update(self, instance, validated_data):

        if instance.info is not None and validated_data.get('info') is not None:
            new_info = JSONField().to_representation(instance.info)
            new_info.update(validated_data['info'])
            validated_data['info'] = new_info

        return super().update(instance, validated_data)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Alert message model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .activity import Activity
from .assessment_session import AssessmentSession
from .base_model import BaseModel
from .instrument import Instrument
from .learner import Learner
from .message import Message

ALERT_LEVEL = (
    (0, _('Info')),
    (1, _('Warning')),
    (2, _('Alert')),
    (3, _('Error')),
)


def get_alert_from_value(value):
    """
        Get the alert level integer value from an string
        :param value: String with the alert level
        :type value: str
        :return: Alert level ID
        :rtype: int
    """
    if value.upper() == 'INFO':
        return 0
    if value.upper() == 'WARNING':
        return 1
    if value.upper() == 'ALERT':
        return 2
    if value.upper() == 'ERROR':
        return 3
    return None


class Alert(BaseModel):
    """ Alert message model. """
    learner = models.ForeignKey(Learner, null=False,
                                on_delete=models.CASCADE,
                                help_text=_("Related learner"))

    level = models.IntegerField(choices=ALERT_LEVEL, null=False,
                                default=0,
                                help_text=_('Level of the alert'))

    instruments = models.ManyToManyField(Instrument, help_text=_("Instruments related to this alert"))

    session = models.ForeignKey(AssessmentSession, null=True, default=None,
                                on_delete=models.SET_NULL,
                                help_text=_("Related assessment session"))

    activity = models.ForeignKey(Activity, null=False,
                                 on_delete=models.CASCADE,
                                 help_text=_("Related activity"))

    message_code = models.ForeignKey(Message, null=True, default=None,
                                     on_delete=models.SET_NULL,
                                     help_text=_("Related message code"))

    data = models.FileField(max_length=250, null=False, blank=False, help_text=_("Data path on storage."))
    raised_at = models.DateTimeField(null=False, help_text=_('When the alert was raised.'))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __repr__(self):
        return "<Alert(learner='%s', activity='%r', level='%r')>" % (
            self.learner, self.activity, self.get_level_display()
        )



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""User api serialize module."""
from django.contrib.auth.models import make_password
from rest_framework import serializers
from rest_framework import validators

from tesla_ce.models import User
from tesla_ce.models import InstitutionUser
from tesla_ce import get_default_client


class UserSerializer(serializers.ModelSerializer):
    """User serialize model module."""

    username = serializers.CharField(allow_blank=False,
                                     allow_null=False,
                                     validators=[
                                         validators.UniqueValidator(queryset=User.objects.all())
                                     ]
                                     )
    last_login = serializers.DateTimeField(read_only=True)
    first_name = serializers.CharField(required=False, default=None)
    last_name = serializers.CharField(required=False, default=None)
    email = serializers.EmailField(required=False, default=None,
                                   validators=[
                                       validators.UniqueValidator(queryset=User.objects.all())
                                   ]
                                   )
    password = serializers.CharField(write_only=True, default=None, allow_null=True)
    password2 = serializers.CharField(write_only=True, default=None, allow_null=True)
    institution = serializers.SerializerMethodField(read_only=True)
    roles = serializers.SerializerMethodField(read_only=True)
    institution_id = serializers.IntegerField(write_only=True, allow_null=True, default=None)
    login_allowed = serializers.BooleanField(required=False, default=True, allow_null=False)
    uid = serializers.CharField(write_only=True, default=None, allow_null=True)

    is_staff = serializers.BooleanField(required=False, allow_null=False, default=False)
    inst_admin = serializers.BooleanField(required=False, allow_null=False, default=False)
    legal_admin = serializers.BooleanField(required=False, allow_null=False, default=False)
    send_admin = serializers.BooleanField(required=False, allow_null=False, default=False)
    data_admin = serializers.BooleanField(required=False, allow_null=False, default=False)

    class Meta:
        model = User
        fields = '__all__'

    def validate(self, attrs):
        """
            Validate the given attributes
            :param attrs: Attributes parsed from request
            :type attrs: dict
            :return: Validated attributes
            :rtype: dict
        """
        # Check passwords
        if attrs.get('login_allowed', True):
            if 'password' in attrs and attrs.get('password') != attrs.get('password2'):
                raise serializers.ValidationError('Passwords does not match')
            if 'password2' in attrs:
                del attrs['password2']
        else:
            attrs['login_allowed'] = False

        if 'institution_id' in attrs and attrs['institution_id'] == -1:
            attrs['institution_id'] = None

        if (self.instance is None or self.get_institution(self.instance) is None) and \
                attrs.get('institution_id') is not None and attrs.get('uid') is None:
            raise ValueError('uid value is required when assigning an institution')

        return super().validate(attrs)

    def get_institution(self, object):
        if hasattr(object, "institutionuser"):
            roles = []
            if hasattr(object.institutionuser, "instructor"):
                roles.append("INSTRUCTOR")
            if hasattr(object.institutionuser, "learner"):
                roles.append("LEARNER")
            if object.institutionuser.inst_admin:
                roles.append("ADMIN")
            if object.institutionuser.send_admin:
                roles.append("SEND")
            if object.institutionuser.legal_admin:
                roles.append("LEGAL")
            if object.institutionuser.data_admin:
                roles.append("DATA")

            return {
                "id": object.institutionuser.institution.id,
                "acronym": object.institutionuser.institution.acronym,
                "uid": object.institutionuser.uid,
                "roles": roles,
                "locale": object.institutionuser.locale,
            }

        return None

    def get_roles(self, object):
        roles = []
        inst = self.get_institution(object)
        if inst is not None:
            roles = inst['roles']
        if object.is_staff:
            roles.append('GLOBAL_ADMIN')

        return roles

    def create(self, validated_data):
        inst_data = None
        if 'institution_id' in validated_data and validated_data['institution_id'] is not None:
            inst_data = {
                'institution': validated_data['institution_id'],
                'uid': None,
                'inst_admin': False,
                'data_admin': False,
                'send_admin': False,
                'legal_admin': False,
                'login_allowed': False
            }
            if 'inst_admin' in validated_data and validated_data['inst_admin'] is not None:
                inst_data['inst_admin'] = validated_data['inst_admin']
            if 'data_admin' in validated_data and validated_data['data_admin'] is not None:
                inst_data['data_admin'] = validated_data['data_admin']
            if 'send_admin' in validated_data and validated_data['send_admin'] is not None:
                inst_data['send_admin'] = validated_data['send_admin']
            if 'legal_admin' in validated_data and validated_data['legal_admin'] is not None:
                inst_data['legal_admin'] = validated_data['legal_admin']
            if 'uid' in validated_data and validated_data['uid'] is not None:
                inst_data['uid'] = validated_data['uid']
            if 'login_allowed' in validated_data and validated_data['login_allowed'] is not None:
                inst_data['login_allowed'] = validated_data['login_allowed']
        if 'institution_id' in validated_data:
            del validated_data['institution_id']
        if 'inst_admin' in validated_data:
            del validated_data['inst_admin']
        if 'legal_admin' in validated_data:
            del validated_data['legal_admin']
        if 'data_admin' in validated_data:
            del validated_data['data_admin']
        if 'send_admin' in validated_data:
            del validated_data['send_admin']
        if 'uid' in validated_data:
            del validated_data['uid']
        if 'login_allowed' in validated_data:
            del validated_data['login_allowed']

        new_password = None
        if 'password' in validated_data:
            new_password = validated_data['password']
            del validated_data['password']

        user = super().create(validated_data)

        if inst_data is not None:
            # Disable password if necessary
            if not inst_data['login_allowed']:
                user.set_unusable_password()
                user.save()
                new_password = None

            # Create the related institution user
            user.institutionuser = InstitutionUser(institution_id=inst_data['institution'],
                                                   uid=inst_data['uid'],
                                                   login_allowed=inst_data['login_allowed'],
                                                   inst_admin=inst_data['inst_admin'],
                                                   legal_admin=inst_data['legal_admin'],
                                                   data_admin=inst_data['data_admin'],
                                                   send_admin=inst_data['send_admin'])
            user.institutionuser.save()
            user.save()
            user.institutionuser.refresh_from_db()

        if new_password:
            get_default_client().change_user_password(user.email, new_password)
            user.refresh_from_db()

        return user

    def update(self, instance, validated_data):
        has_inst = self.get_institution(instance) is not None
        new_inst = False

        # Check if we are adding a new institution
        if 'institution_id' in validated_data:
            if not has_inst and validated_data['institution_id'] is not None:
                # Create the related institution user
                instance.institutionuser = InstitutionUser(institution_id=validated_data['institution_id'],
                                                           uid=None,
                                                           login_allowed=False,
                                                           inst_admin=False,
                                                           legal_admin=False,
                                                           data_admin=False,
                                                           send_admin=False)
                new_inst = True
            elif has_inst and validated_data['institution_id'] is None:
                # Removing the institution
                has_inst = False
                instance.institutionuser.delete(keep_parents=True)
                instance.refresh_from_db()

        # Apply institution allowed fields modifications
        if has_inst or new_inst:
            modified = False
            if 'inst_admin' in validated_data:
                instance.institutionuser.inst_admin = validated_data['inst_admin']
                modified = True
            if 'send_admin' in validated_data:
                instance.institutionuser.send_admin = validated_data['send_admin']
                modified = True
            if 'legal_admin' in validated_data:
                instance.institutionuser.legal_admin = validated_data['legal_admin']
                modified = True
            if 'data_admin' in validated_data:
                instance.institutionuser.data_admin = validated_data['data_admin']
                modified = True
            if 'uid' in validated_data:
                instance.institutionuser.uid = validated_data['uid']
                modified = True
            if 'login_allowed' in validated_data:
                instance.institutionuser.login_allowed = validated_data['login_allowed']
                modified = True
            if 'institution_id' in validated_data:
                instance.institutionuser.institution_id = validated_data['institution_id']
                modified = True

            if modified or new_inst:
                instance.institutionuser.save()

        # If a new institution has been added, refresh objects to avoid data loose
        if new_inst:
            instance.save()
            instance.institutionuser.refresh_from_db()

        new_password = None
        if 'password' in validated_data:
            new_password = validated_data['password']
            del validated_data['password']

        updated_user = super().update(instance, validated_data)

        if new_password is not None:
            get_default_client().change_user_password(updated_user.email, new_password)

        return updated_user



# Generated by Django 3.0.8 on 2020-07-12 11:30

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0006_auto_20200520_2128'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='enrolment',
            name='model_samples',
        ),
        migrations.AddField(
            model_name='enrolment',
            name='model_samples',
            field=models.ManyToManyField(help_text='Enrolment samples used to create this model', related_name='models', to='tesla_ce.EnrolmentSample'),
        ),
    ]



# Generated by Django 3.2.4 on 2021-10-09 00:08

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0020_institution_allowed_domains'),
    ]

    operations = [
        migrations.AlterField(
            model_name='reportactivitysession',
            name='session',
            field=models.OneToOneField(help_text='Assessment Session related to this report.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.assessmentsession'),
        ),
    ]



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" SENDCategory model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .institution import Institution


class SENDCategory(BaseModel):
    """ SENDCategory model. """
    institution = models.ForeignKey(Institution, null=False, on_delete=models.CASCADE,
                                    help_text=_('Institution this category is defined for'))
    description = models.TextField(null=False, blank=False,
                                   help_text=_("Category description."))

    data = models.TextField(null=False, blank=False,
                            help_text=_("Category configuration."))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __repr__(self):
        return "<SENDCategory(id='%r', description='%s')>" % (
            self.id, self.description)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Activity Report views module """
from rest_framework import viewsets
from rest_framework import exceptions
from rest_framework_extensions.mixins import NestedViewSetMixin

from tesla_ce.apps.api import permissions
from tesla_ce.apps.api.v2.serializers import InstitutionCourseActivityReportAuditSerializer

from tesla_ce.models import Course
from tesla_ce.models import ReportActivityInstrument
from tesla_ce.models.user import get_institution_user
from tesla_ce.models.user import is_global_admin


# pylint: disable=too-many-ancestors
class InstitutionCourseActivityReportAuditViewSet(NestedViewSetMixin, viewsets.ReadOnlyModelViewSet):
    """
    API endpoint that allows view activity report audit data.
    """
    model = ReportActivityInstrument
    serializer_class = InstitutionCourseActivityReportAuditSerializer
    lookup_field = 'instrument_id'
    permission_classes = [
        permissions.GlobalAdminReadOnlyPermission |
        permissions.InstitutionAdminReadOnlyPermission |
        permissions.InstitutionDataAdminReadOnlyPermission |
        permissions.InstitutionCourseInstructorReadOnlyPermission |
        permissions.InstitutionCourseLearnerReadOnlyPermission
    ]

    def get_queryset(self):
        queryset = self.filter_queryset_by_parents_lookups(ReportActivityInstrument.objects)
        if not is_global_admin(self.request.user):
            inst_user = get_institution_user(self.request.user)
            if not inst_user.inst_admin and not inst_user.legal_admin and not inst_user.data_admin:
                course = Course.objects.get(
                    vle__institution=self.kwargs['parent_lookup_report__activity__vle__institution_id'],
                    id=self.kwargs['parent_lookup_report__activity__course_id']
                )
                try:
                    is_instructor = self.request.user.instructor in course.instructors.all()
                except Exception:
                    is_instructor = False
                if not is_instructor:
                    if not inst_user.institution.allow_learner_report or not inst_user.institution.allow_learner_audit:
                        raise exceptions.PermissionDenied('Not allowed')
                    queryset = queryset.filter(report__learner_id=inst_user.id)
        return queryset.all()



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Course api serialize module """
from rest_framework import serializers

from tesla_ce.models import Course
from .vle import VLESerializer


class VLECourseSerializer(serializers.ModelSerializer):
    """Course serialize class."""

    vle = VLESerializer(read_only=True)
    vle_id = serializers.HiddenField(default=None, allow_null=True)
    enabled = serializers.BooleanField(read_only=True)

    class Meta:
        model = Course
        exclude = ["learners", "instructors"]
        validators = [serializers.UniqueTogetherValidator(
            queryset=Course.objects.all(),
            fields=['vle_id', 'vle_course_id']
        ), ]

    def validate(self, attrs):
        """
            Validate the given attributes
            :param attrs: Attributes parsed from request
            :type attrs: dict
            :return: Validated attributes
            :rtype: dict
        """
        # Add predefined fields
        attrs['vle_id'] = self.context['view'].kwargs['parent_lookup_vle_id']

        # Apply validators
        for validator in self.get_validators():
            validator(attrs, self)
        return super().validate(attrs)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
""" Vault unseal script management tests """
import os
import pytest
import mock

from io import StringIO

from django.core.management import call_command


def test_unseal(tesla_ce_system):

    #pytest.skip("Not valid with Vault in dev mode. TODO: Start test service for Vault in production mode")

    assert tesla_ce_system is not None

    out = StringIO()
    err = StringIO()

    # If local configuration file is not available, generate configuration
    if not os.path.exists('tesla-ce.cfg'):
        # Write the configuration file to disk
        with open('tesla-ce.cfg', 'w') as out_fh:
            tesla_ce_system.config.config.write(out_fh)
    with mock.patch('tesla_ce.management.base.TeslaCommand.get_client', return_value=tesla_ce_system):
        call_command(
            'unseal',
            stdout=out,
            stderr=err,
            local=True
        )



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Institution views definition package"""
from .activity import InstitutionCourseActivityViewSet
from .activity_instrument import InstitutionCourseActivityInstrumentViewSet
from .activity_report import InstitutionCourseActivityReportViewSet
from .activity_report_audit import InstitutionCourseActivityReportAuditViewSet
from .course import InstitutionCourseViewSet
from .course_learner import InstitutionCourseLearnerViewSet
from .course_instructor import InstitutionCourseInstructorViewSet
from .course_group import InstitutionCourseGroupCourseViewSet
from .course_group import InstitutionCourseGroupViewSet
from .informed_consent import InstitutionInformedConsentViewSet
from .informed_consent_document import InstitutionInformedConsentDocumentViewSet
from .institution import InstitutionViewSet
from .instructor import InstitutionInstructorViewSet
from .instrument import InstitutionInstrumentViewSet
from .learner import InstitutionLearnerViewSet
from .request import InstitutionCourseActivityReportRequestViewSet
from .send_category import InstitutionSENDCategoryViewSet
from .send_learner import InstitutionSENDLearnerViewSet
from .ui_option import InstitutionUIOptionViewSet
from .user import InstitutionUserViewSet
from .vle import InstitutionVLEViewSet

__all__ = [
    "InstitutionViewSet",
    "InstitutionCourseActivityViewSet",
    "InstitutionCourseActivityReportViewSet",
    "InstitutionCourseActivityInstrumentViewSet",
    "InstitutionCourseViewSet",
    "InstitutionCourseGroupViewSet",
    "InstitutionCourseGroupCourseViewSet",
    "InstitutionLearnerViewSet",
    "InstitutionCourseActivityReportRequestViewSet",
    "InstitutionCourseActivityReportAuditViewSet",
    "InstitutionCourseLearnerViewSet",
    "InstitutionCourseInstructorViewSet",
    "InstitutionInstructorViewSet",
    "InstitutionInformedConsentViewSet",
    "InstitutionInformedConsentDocumentViewSet",
    "InstitutionSENDCategoryViewSet",
    "InstitutionSENDLearnerViewSet",
    "InstitutionVLEViewSet",
    "InstitutionUIOptionViewSet",
    "InstitutionUserViewSet",
    "InstitutionInstrumentViewSet",
]



#  Copyright (c) 2021 Mireia Bellot
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Test module for institution course groups management """
import logging

import pytest

import tests.utils


@pytest.mark.django_db
def test_api_institution_course_groups(rest_api_client, user_global_admin, institution_course_test_case):
    institution_user = institution_course_test_case['user'].institutionuser
    institution = institution_course_test_case['institution']
    institution_id = institution.id

    user_global_admin.is_staff = True
    user_global_admin.save()

    # Set global admin user.
    rest_api_client.force_authenticate(user=user_global_admin)

    # Institution Admin privileges
    institution_user.inst_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    # List Groups
    logging.info('\n1) LIST GROUPS --------------------------------------')
    str_path = '/api/v2/institution/{}/group/'.format(institution_id)
    str_response = 'RESPONSE id={}:'.format(institution_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Groups', str_response, 200)
    # Ensure that no group exists
    assert body['count'] == 0

    # Create a new Group
    logging.info('\n2) CREATE A NEW GROUP --------------------------------------')
    str_data = {'parent': '', 'name': 'TEST_GROUP', 'description': 'This is a test group'}
    new_group_id = tests.utils.post_rest_api_client(rest_api_client, str_path, str_data,
                                                    'Create a new group', 'RESPONSE: ', 201)

    # List groups
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Groups', 'RESPONSE:', 200)

    # Ensure group exists
    n_groups = body['count']
    assert n_groups == 1

    # Create a new nested Group
    logging.info('\n2) CREATE A NEW NESTED GROUP --------------------------------------')
    str_data = {'parent': new_group_id , 'name': 'TEST_GROUP', 'description': 'This is a test group'}
    new_subgroup_id = tests.utils.post_rest_api_client(rest_api_client, str_path, str_data,
                                                       'Create a new nested group', 'RESPONSE: ', 201)

    # List groups
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Groups', 'RESPONSE:', 200)

    # Ensure group exists
    n_groups = body['count']
    assert n_groups == 2

    # Filter root groups (parent is None)
    body = tests.utils.get_rest_api_client(rest_api_client, '{}?parent={}'.format(str_path, -1),
                                           'List Root Groups', 'RESPONSE:', 200)
    assert body['count'] == 1
    assert body['results'][0]['id'] == new_group_id

    # Filter child groups
    body = tests.utils.get_rest_api_client(rest_api_client, '{}?parent={}'.format(str_path, new_group_id),
                                           'List Child Groups', 'RESPONSE:', 200)
    assert body['count'] == 1
    assert body['results'][0]['id'] == new_subgroup_id


    # Read group information
    logging.info('\n3) READ GROUP INFORMATION --------------------------------------')
    str_path = '/api/v2/institution/{}/group/{}/'.format(institution_id, new_group_id)
    str_response = 'RESPONSE group id={}:'.format(new_group_id)
    tests.utils.get_rest_api_client(rest_api_client, str_path,
                                    'Read Group Information', str_response, 200)

    # Update group data
    logging.info('\n4) UDPATE GROUP INFORMATION --------------------------------------')
    str_data = {'name': 'CHANGED_TEST_GROUP', 'description': 'This is a CHANGED TEXT FOR THE test group'}

    tests.utils.put_rest_api_client(rest_api_client, str_path, str_data,
                                    'Update group', 'RESPONSE: ', 200)

    # Delete root group
    logging.info('\n5) DELETE ROOT GROUP --------------------------------------')
    str_path = '/api/v2/institution/{}/group/{}/'.format(institution_id, new_group_id)
    tests.utils.delete_rest_api_client(rest_api_client, str_path,
                                       'Delete group', "RESPONSE: ", 204)
    str_path = '/api/v2/institution/{}/group/'.format(institution_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Groups', 'RESPONSE:', 200)
    logging.info(body)

    # Ensure that the root group and child group are removed
    assert body['count'] == 0


@pytest.mark.django_db
def test_api_institution_course_group_courses(rest_api_client, user_global_admin, institution_course_test_case):
    institution_user = institution_course_test_case['user'].institutionuser
    institution = institution_course_test_case['institution']
    institution_id = institution.id

    user_global_admin.is_staff = True
    user_global_admin.save()

    # Set global admin user.
    rest_api_client.force_authenticate(user=user_global_admin)

    # Institution Admin privileges
    institution_user.inst_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    # Create a new Group
    logging.info('\n2) CREATE A NEW GROUP --------------------------------------')
    str_path = '/api/v2/institution/{}/group/'.format(institution_id)
    str_data = {'parent': '', 'name': 'TEST_GROUP', 'description': 'This is a test group'}
    new_group_id = tests.utils.post_rest_api_client(rest_api_client, str_path, str_data,
                                                    'Create a new group', 'RESPONSE: ', 201)

    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Groups', 'RESPONSE:', 200)
    assert body['count'] == 1

    # List courses in a group
    logging.info('\n6) LIST COURSES IN A GROUP --------------------------------------')
    str_path = '/api/v2/institution/{}/group/{}/course/'.format(institution_id, new_group_id)

    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Courses in a Group', 'RESPONSE:', 200)

    # Ensure no courses in a new group by default
    assert body['count'] == 0

    # Add a new course
    logging.info('\n7) ADD A COURSE TO A GROUP --------------------------------------')
    str_path = '/api/v2/institution/{}/group/{}/course/'.format(institution_id, new_group_id)
    course_id = institution_course_test_case['course'].id
    tests.utils.post_rest_api_client(rest_api_client, str_path, {'id': course_id},
                                     'Add a Course to a Group', 'RESPONSE: ', 201)

    # List and check number of courses in a group has increased.
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Courses in a Group', 'RESPONSE:', 200)
    assert body['count'] == 1
    assert body['results'][0]['id'] == course_id

    # Delete a course from a group
    logging.info('\n8) DELETE A COURSE FROM A GROUP --------------------------------------')
    str_path_course = '/api/v2/institution/{}/group/{}/course/{}/'.format(institution_id, new_group_id, course_id)
    tests.utils.delete_rest_api_client(rest_api_client, str_path_course,
                                       'Delete a Course from a Group', "RESPONSE: ", 204)
    tests.utils.get_rest_api_client(rest_api_client, str_path_course,
                                    'Show removed course', 'RESPONSE:', 404)

    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List Courses in a Group', 'RESPONSE:', 200)
    assert body['count'] == 0



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" API views package """
from .admin import AdminProviderViewSet
from .admin import AdminUserViewSet
from .admin import InstitutionAdminViewSet
from .admin import InstrumentViewSet
from .admin import UIOptionViewSet
from .institution import InstitutionCourseActivityInstrumentViewSet
from .institution import InstitutionCourseActivityReportViewSet
from .institution import InstitutionCourseActivityReportAuditViewSet
from .institution import InstitutionCourseActivityViewSet
from .institution import InstitutionCourseGroupCourseViewSet
from .institution import InstitutionCourseGroupViewSet
from .institution import InstitutionCourseViewSet
from .institution import InstitutionInformedConsentDocumentViewSet
from .institution import InstitutionInformedConsentViewSet
from .institution import InstitutionInstructorViewSet
from .institution import InstitutionInstrumentViewSet
from .institution import InstitutionLearnerViewSet
from .institution import InstitutionCourseLearnerViewSet
from .institution import InstitutionCourseInstructorViewSet
from .institution import InstitutionCourseActivityReportRequestViewSet
from .institution import InstitutionSENDCategoryViewSet
from .institution import InstitutionSENDLearnerViewSet
from .institution import InstitutionUIOptionViewSet
from .institution import InstitutionUserViewSet
from .institution import InstitutionVLEViewSet
from .institution import InstitutionViewSet
from .provider import ProviderEnrolmentSampleValidationViewSet
from .provider import ProviderEnrolmentSampleViewSet
from .provider import ProviderEnrolmentViewSet
from .provider import ProviderNotificationViewSet
from .provider import ProviderVerificationRequestResultViewSet
from .provider import ProviderViewSet
from .vle import VLECourseActivityInstrumentViewSet
from .vle import VLECourseActivityLearnerRequestViewSet
from .vle import VLECourseActivityLearnerViewSet
from .vle import VLECourseActivityReportViewSet
from .vle import VLECourseActivityViewSet
from .vle import VLECourseInstructorViewSet
from .vle import VLECourseLearnerViewSet
from .vle import VLECourseViewSet
from .vle import VLEInstrumentViewSet
from .vle import VLEViewSet

__all__ = [
    # Administration nested views
    "InstitutionAdminViewSet",
    "InstrumentViewSet",
    "AdminProviderViewSet",
    "AdminUserViewSet",
    "UIOptionViewSet",
    # VLE nested views
    "VLEViewSet",
    "VLECourseViewSet",
    "VLECourseActivityViewSet",
    "VLECourseActivityInstrumentViewSet",
    "VLECourseActivityReportViewSet",
    "VLECourseLearnerViewSet",
    "VLECourseActivityLearnerViewSet",
    "VLECourseActivityLearnerRequestViewSet",
    "VLECourseInstructorViewSet",
    "VLEInstrumentViewSet",
    # Institution nested views
    "InstitutionViewSet",
    "InstitutionCourseViewSet",
    "InstitutionCourseGroupViewSet",
    "InstitutionCourseGroupCourseViewSet",
    "InstitutionCourseActivityViewSet",
    "InstitutionCourseActivityReportViewSet",
    "InstitutionCourseActivityReportAuditViewSet",
    "InstitutionCourseActivityInstrumentViewSet",
    "InstitutionLearnerViewSet",
    "InstitutionCourseLearnerViewSet",
    "InstitutionCourseInstructorViewSet",
    "InstitutionCourseActivityReportRequestViewSet",
    "InstitutionInstructorViewSet",
    "InstitutionInformedConsentViewSet",
    "InstitutionInformedConsentDocumentViewSet",
    "InstitutionSENDCategoryViewSet",
    "InstitutionSENDLearnerViewSet",
    "InstitutionVLEViewSet",
    "InstitutionUIOptionViewSet",
    "InstitutionUserViewSet",
    # Provider nested views
    "ProviderViewSet",
    "ProviderEnrolmentViewSet",
    "ProviderEnrolmentSampleViewSet",
    "ProviderEnrolmentSampleValidationViewSet",
    "ProviderVerificationRequestResultViewSet",
    "ProviderNotificationViewSet",
]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Notifications for providers model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .provider import Provider


class ProviderNotification(BaseModel):
    """ Provider notification model. """
    provider = models.ForeignKey(Provider, null=False,
                                on_delete=models.CASCADE,
                                help_text=_("Related provider"))

    key = models.CharField(max_length=250, null=False,
                           help_text=_("Notification unique key for the provider"))

    info = models.TextField(null=True, blank=False, help_text=_("Notification information."))
    when = models.DateTimeField(null=False, help_text=_('When the notification should be sent to provider.'))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (('provider', 'key'),)

    def __repr__(self):
        return "<ProviderNotification(provider='%s', key='%r', when='%r')>" % (
            self.provider, self.key, self.info
        )



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Django API URLs definition """
from django.urls import include
from django.urls import path
from django.urls import re_path
from drf_yasg import openapi
from drf_yasg.views import get_schema_view
from rest_framework import permissions

schema_view = get_schema_view(
   openapi.Info(
      title="TeSLA CE API",
      default_version='v2',
      description="TeSLA CE API",
      terms_of_service="https://www.gnu.org/licenses/agpl-3.0.html",
      contact=openapi.Contact(email="xbaro@uoc.edu"),
      license=openapi.License(name="AGPLv3 License"),
   ),
   public=True,
   permission_classes=(permissions.AllowAny,),
)

urlpatterns = [
    path('v1/', include('tesla_ce.apps.api.v1.urls', namespace='v1')),
    path('v2/', include('tesla_ce.apps.api.v2.urls', namespace='v2')),
    re_path(r'swagger(?P<format>\.json|\.yaml)$', schema_view.without_ui(cache_timeout=0), name='schema-json'),
    path(r'swagger/', schema_view.with_ui('swagger', cache_timeout=0), name='schema-swagger-ui'),
    path(r'redoc/', schema_view.with_ui('redoc', cache_timeout=0), name='schema-redoc'),
]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.

from .activity import VLECourseActivityViewSet
from .activity_instrument import VLECourseActivityInstrumentViewSet
from .activity_report import VLECourseActivityReportViewSet
from .course import VLECourseViewSet
from .instructor import VLECourseInstructorViewSet
from .instrument import VLEInstrumentViewSet
from .learner import VLECourseActivityLearnerViewSet
from .learner import VLECourseLearnerViewSet
from .request import VLECourseActivityLearnerRequestViewSet
from .vle import VLEViewSet

__all__ = [
    "VLEViewSet",
    "VLECourseViewSet",
    "VLECourseActivityViewSet", "VLECourseActivityInstrumentViewSet", "VLECourseActivityReportViewSet",
    "VLECourseLearnerViewSet",
    "VLECourseActivityLearnerViewSet",
    "VLECourseInstructorViewSet",
    "VLECourseActivityLearnerRequestViewSet",
    "VLEInstrumentViewSet",
]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Activity serializer module """
from rest_framework import serializers

from tesla_ce.models import Activity

from .course import InstitutionCourseSerializer
from .activity_instrument import InstitutionCourseActivityInstrumentSerializer


class InstitutionCourseActivitySerializer(serializers.ModelSerializer):
    """Activity serialize model module."""
    course_id = serializers.ReadOnlyField()
    vle_id = serializers.ReadOnlyField()
    vle_activity_type = serializers.ReadOnlyField()
    vle_activity_id = serializers.ReadOnlyField()
    name = serializers.ReadOnlyField()
    start = serializers.ReadOnlyField()
    end = serializers.ReadOnlyField()
    description = serializers.ReadOnlyField()

    class Meta:
        model = Activity
        exclude = ["vle", "course"]


class InstitutionCourseActivityExtendedSerializer(serializers.ModelSerializer):
    """ Activity extended serialize model module. """
    vle_id = serializers.ReadOnlyField()
    vle_activity_type = serializers.ReadOnlyField()
    vle_activity_id = serializers.ReadOnlyField()
    name = serializers.ReadOnlyField()
    start = serializers.ReadOnlyField()
    end = serializers.ReadOnlyField()
    description = serializers.ReadOnlyField()
    course = InstitutionCourseSerializer(read_only=True)
    instruments = InstitutionCourseActivityInstrumentSerializer(source='configuration', read_only=True, many=True)
    user_instruments = serializers.SerializerMethodField()

    class Meta:
        model = Activity
        exclude = ["vle"]

    def get_user_instruments(self, instance):
        """ Instruments for current user """
        learner = self.context['request'].user.learner
        return InstitutionCourseActivityInstrumentSerializer(
            instance.get_learner_instruments(learner), many=True).data



#  Copyright (c) 2021 Mireia Bellot
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Tests for institution administration """
import logging

import pytest

import tests.utils


##############################################################################
#
#   ADMIN TESTING: INSTITUTIONS

#   1) List institutions --> 2) Read 1st institution --> 3) Read non existing institution
#   4) Create institution --> 5) List institutions (check +1) --> 6) Read institution --> 7) failed create institution
#   8) update institution --> 9) read institution --> 10) delete institution --> 11) list institution (check it's -1)
#
##############################################################################
@pytest.mark.django_db
def test_api_admin_institutions(rest_api_client, user_global_admin):
    # TODO: add list of not allowed users as 3rd parameter to check_403 function
    # Check that unauthenticated calls are not allowed
    user_global_admin.is_staff = False
    user_global_admin.save()
    tests.utils.check_403(rest_api_client, '/api/v2/admin/institution/', [user_global_admin])

    # Authorization – JWT with Global Administration privileges
    # Set global admin user.
    user_global_admin.is_staff = True
    user_global_admin.save()
    rest_api_client.force_authenticate(user=user_global_admin)

    # 666 TODO: list institutions to DOC
    logging.info('\n1) LIST INSTITUTIONS --------------------------------------')
    body = tests.utils.get_rest_api_client(rest_api_client, '/api/v2/admin/institution/',
                                           'List institutions', 'RESPONSE:', 200)

    n_institution = body['count']
    institutions_empty = True
    id_non_existing_institution = 1000000
    if n_institution > 0:
        institutions_empty = False
        id_first_institution = body['results'][0]['id']
        tests.utils.check_pagination(rest_api_client, body)
        while body['next'] is not None:
            body = tests.utils.get_rest_api_client(rest_api_client, body['next'], 'List institutions', 'Pagination',
                                                   200)
        id_last_institution = body['results'][len(body['results']) - 1]['id']
    else:
        id_first_institution = -1
        id_last_institution = -1

    tests.utils.print_log([['List Institution'], ['NUMBER OF INSTITUTIONS: ', n_institution],
                           ['ID FIRST INSTITUTION: ', id_first_institution],
                           ['ID LAST INSTITUTION: ', id_last_institution],
                           ['ID NON EXISTING INSTITUTION: ', id_non_existing_institution]])

    # READ existing 1st institution
    logging.info('\n2) READ 1st INSTITUTION --------------------------------------')

    # 666 TODO: READ institutions to DOC
    if not institutions_empty:
        str_response = 'RESPONSE id=' + str(id_first_institution) + ':'
        tests.utils.get_rest_api_client(rest_api_client,
                                        '/api/v2/admin/institution/' + str(id_first_institution) + '/',
                                        'Read Institution Information', str_response, 200)

    # READ non existing institution with numeric id
    str_url = '/api/v2/admin/institution/' + str(id_non_existing_institution) + '/'
    str_response = 'RESPONSE id=' + str(id_non_existing_institution) + ':'
    tests.utils.get_rest_api_client(rest_api_client, str_url,
                                    'Read Institution Information', str_response, 404)
    # -----------------------------------------------------------------------------
    # Create a new institution:
    #   POST /api/v2/admin/institution/
    #   Status Codes:
    #       201 Created – Created
    #       400 Bad Request – Invalid information provided. The response contains the description of the errors.
    # -----------------------------------------------------------------------------
    # CREATE new institution: SUCCESS
    logging.info('\n4) CREATE INSTITUTION --------------------------------------')
    # 666 TODO: Include mandatory fields and default values to DOC for CREATE NEW INSTITUTION
    """ 666
    acronym(string) MANDATORY field.
    name(string)  MANDATORY field
    external_ic(boolean): False by default.
    mail_domain(string): None by default.
    disable_vle_learner_creation(boolean): False by default.
    disable_vle_instructor_creation(boolean): False by default.
    disable_vle_user_creation(boolean): False by default.
    """
    institution = {'acronym': 'nni', 'name': 'name_new_institution'}
    new_institution_id = tests.utils.post_rest_api_client(rest_api_client, '/api/v2/admin/institution/',
                                                          institution, 'Create New Institution', 'RESPONSE:', 201)

    # 666 TODO: trobar paràmetres obligatoris i opcionals de create institution

    # LIST institution and check count has increased.
    logging.info('\n5) LIST INSTITUTIONS --------------------------------------')
    aux_body = tests.utils.get_rest_api_client(rest_api_client, '/api/v2/admin/institution/',
                                               'Create New Institution', 'RESPONSE:', 200)
    # Get number of institutions
    aux_n_institution = aux_body['count']

    args = [['Create New Institution'], ['Number of Institutions PRE: ', n_institution],
            ['Number of Institutions POST:', aux_n_institution]]

    tests.utils.print_log(args)

    assert aux_n_institution == n_institution + 1

    # READ new institution
    logging.info('\n6) READ NEW INSTITUTION --------------------------------------')
    str_url = '/api/v2/admin/institution/' + str(new_institution_id) + '/'
    aux_body = tests.utils.get_rest_api_client(rest_api_client, str_url,
                                               'Create New Institution [Read Institution Information]',
                                               'RESPONSE id=', 200)
    args = [['Create New Institution [Read Institution Information]'], ['RESPONSE id=: ', new_institution_id],
            ['', aux_body]]

    tests.utils.print_log(args)
    print('Create New Institution [Read Institution Information] RESPONSE id=:', new_institution_id, ' ', aux_body)

    # CREATE new institution: FAILED
    logging.info('\n7) FAILED CREATION NEW INSTITUTION --------------------------------------')
    failed_institution = {'acronym': 'n3'}
    tests.utils.post_rest_api_client(rest_api_client, '/api/v2/admin/institution/',
                                     failed_institution, 'Create New Institution',
                                     'RESPONSE:', 400)

    # -----------------------------------------------------------------------------
    # Update institution information:
    #   PUT /api/v2/admin/institution/(int: institution_id)/
    #   Status Codes:
    #       200 OK – Ok
    #       400 Bad Request – Invalid information provided. The response contains the description of the errors.
    #       404 Not Found – Institution not found
    # -----------------------------------------------------------------------------
    # [Status: 200] UPDATE new institution
    logging.info('\n8) UPDATE INSTITUTION --------------------------------------')
    str_data = {'acronym': 'CHANGED', 'name': 'name3', 'queue': 'tesla_test23'}
    str_path = '/api/v2/admin/institution/' + str(new_institution_id) + '/'
    tests.utils.put_rest_api_client(rest_api_client, str_path, str_data, 'Update institution information',
                                    'RESPONSE:', 200)
    # [Status: 400] UPDATE WRONG NUMBER PARAMETERS. REQUIRED: name, queue
    str_data = {'acronym': 'CHANGED'}
    tests.utils.put_rest_api_client(rest_api_client, str_path, str_data, 'Update institution information',
                                    'RESPONSE:', 400)

    # [Status: 404] UPDATE NON EXISTING institution
    str_path = '/api/v2/admin/institution/' + str(new_institution_id + 1) + '/'
    tests.utils.put_rest_api_client(rest_api_client, str_path, str_data, 'Update institution information',
                                    'RESPONSE:', 404)

    # Read updated institution
    logging.info('\n9) READ UPDATED INSTITUTION --------------------------------------')
    str_path = '/api/v2/admin/institution/' + str(new_institution_id) + '/'
    str_response = 'RESPONSE id=' + str(new_institution_id) + ':'
    tests.utils.get_rest_api_client(rest_api_client, str_path,
                                    'Read Updated Institution', str_response, 200)
    """-----------------------------------------------------------------------------
     Delete an institution:
       DELETE /api/v2/admin/institution/(int: institution_id)/
       Status Codes:
           204 No Content – No Content
           404 Not Found – Institution not found
    -----------------------------------------------------------------------------"""
    logging.info('\n10) DELETE UPDATED INSTITUTION --------------------------------------')
    # Delete updated institution
    body = tests.utils.get_rest_api_client(rest_api_client, '/api/v2/admin/institution/',
                                           'List institutions', 'RESPONSE:', 200)
    n_institution_pre = body['count']
    if n_institution_pre > 0:
        id_first_institution_pre = body['results'][0]['id']
        while body['next'] is not None:
            body = tests.utils.get_rest_api_client(rest_api_client, body['next'], 'List institutions', 'Pagination pre',
                                                   200)
        id_last_institution_pre = body['results'][len(body['results']) - 1]['id']
    else:
        id_first_institution_pre = -1
        id_last_institution_pre = -1

    # [Status = 204] Delete existing item
    tests.utils.delete_rest_api_client(rest_api_client, str_path, 'Delete an institution',
                                       'DELETED', 204)

    body = tests.utils.get_rest_api_client(rest_api_client, '/api/v2/admin/institution/',
                                           'List institutions', 'RESPONSE:', 200)
    n_institution_post = body['count']
    if n_institution_post > 0:
        id_first_institution_post = body['results'][0]['id']
        while body['next'] is not None:
            body = tests.utils.get_rest_api_client(rest_api_client, body['next'], 'List institutions',
                                                   'Pagination post', 200)
        id_last_institution_post = body['results'][len(body['results']) - 1]['id']
    else:
        id_first_institution_post = -1
        id_last_institution_post = -1

    str_log = [['Delete an institution'], ['PRE delete: ', n_institution_pre],
               ['items - ID first item:', id_first_institution_pre],
               ['- ID last item:', id_last_institution_pre]]
    tests.utils.print_log(str_log)

    str_log = [['Delete an institution'], ['POST delete: ', n_institution_post],
               ['items - ID first item:', id_first_institution_post],
               ['- ID last item:', id_last_institution_post]]
    tests.utils.print_log(str_log)

    # [Status = 404] DELETE non existing item
    tests.utils.delete_rest_api_client(rest_api_client, str_path, 'Delete an institution',
                                       'NOT DELETED', 404)

    logging.info('\n11) LIST INSTITUTIONS -----------------------------------------------')
    tests.utils.get_rest_api_client(rest_api_client, '/api/v2/admin/institution/',
                                    'List institutions', 'RESPONSE:', 200)



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Learner model module."""
import uuid

from cache_memoize import cache_memoize
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from tesla_ce.apps.api.utils import decode_json
from .informed_consent import InformedConsent
from .user import InstitutionUser


@cache_memoize(24 * 60 * 60)
def get_learner_send(learner_id):
    """
        Compute the learner SEND properties
        :param learner_id: The database id of the learner
        :type learner_id: int
        :return: SEND information for the learner
        :rtype: dict
    """
    learner = Learner.objects.get(id=learner_id)
    send_categories = learner.sendlearner_set.filter(
        models.Q(expires_at__gte=timezone.now()) | models.Q(expires_at__isnull=True)
    ).all()
    disabled_instruments = set()
    options = set()

    if send_categories is None:
        send_categories = []

    for c in send_categories:
        data = decode_json(c.category.data)
        if 'disabled_instruments' in data:
            disabled_instruments.update(data['disabled_instruments'])
        if 'enabled_options' in data:
            options.update(data['enabled_options'])

    return {
        'is_send': len(send_categories) > 0,
        'send': {
            'enabled_options': list(options),
            'disabled_instruments': list(disabled_instruments)
        }
    }


@cache_memoize(24 * 60 * 60)
def get_learner_enrolment(learner_id):
    """
        Compute the enrolment status for a learner
        :param learner_id: The database id of the learner
        :type learner_id: int
        :return: Enrolment status for the learner
        :rtype: dict
    """
    # Get the real enrolment values for instruments and providers
    learner = Learner.objects.get(id=learner_id)
    real_enrolments = learner.enrolment_set.filter(provider__enabled=True).values(
        instrument_id=models.F("provider__instrument_id")
    ).annotate(
        models.Min("percentage"), models.Max("percentage"), models.Min("can_analyse"), models.Max("can_analyse")
    ).all()
    # Compute enrolment estimation from pending enrolment samples
    pending_contribution = learner.enrolmentsample_set.filter(
        enrolmentsamplevalidation__status=1,
        models__isnull=True
    ).values(
        provider_id=models.F("enrolmentsamplevalidation__provider"),
        instrument_id=models.F("enrolmentsamplevalidation__provider__instrument_id"),
    ).annotate(
        pending_contribution=models.Sum("enrolmentsamplevalidation__contribution")
    )
    pending_validation = learner.enrolmentsample_set.filter(
        models.Q(enrolmentsamplevalidation__status=0) | models.Q(enrolmentsamplevalidation__status=4),
        enrolmentsamplevalidation__provider__enabled=True
    ).values(
        provider_id=models.F("enrolmentsamplevalidation__provider"),
        instrument_id=models.F("enrolmentsamplevalidation__provider__instrument_id"),
    ).annotate(
        count=models.Count("id")
    )
    enrolments = list(real_enrolments)
    instruments = []
    for enrol in real_enrolments:
        instruments.append(enrol['instrument_id'])
        pending = pending_contribution.filter(instrument_id=enrol['instrument_id'])
        enrol['pending'] = list(pending.all())
        enrol['not_validated'] = list(pending_validation.all())
        enrol['pending_contributions'] = min(1.0, pending_contribution.filter(
            instrument_id=enrol['instrument_id']).aggregate(models.Avg('pending_contribution')
                                                            )['pending_contribution__avg'] or 0.0)
        enrol['not_validated_count'] = int(pending_validation.filter(
            instrument_id=enrol['instrument_id']).aggregate(models.Avg('count')
                                                            )['count__avg'] or 0)

    missing = {}
    for missing_enrolment in pending_validation.exclude(instrument_id__in=instruments).all():
        pending_val = pending_validation.filter(instrument_id=missing_enrolment['instrument_id'])
        missing[missing_enrolment['instrument_id']] = {
            'instrument_id': missing_enrolment['instrument_id'],
            'percentage__min': 0.0,
            'percentage__max': 0.0,
            'can_analyse__min': False,
            'can_analyse__max': False,
            'pending': [],
            'not_validated': list(pending_val.all()),
            'pending_contributions': 0.0,
            'not_validated_count': int(pending_val.aggregate(models.Avg('count'))['count__avg'] or 0)
        }

    for missing_enrolment in pending_contribution.exclude(instrument_id__in=instruments).all():
        pending_cont = pending_contribution.filter(instrument_id=missing_enrolment['instrument_id'])
        if missing_enrolment['instrument_id'] in missing:
            missing[missing_enrolment['instrument_id']]['pending'] = list(pending_cont.all())
            missing[missing_enrolment['instrument_id']]['pending_contributions'] = min(1.0, pending_cont.aggregate(
                models.Avg('pending_contribution'))['pending_contribution__avg'] or 0.0)
        else:
            missing[missing_enrolment['instrument_id']] = {
                'instrument_id': missing_enrolment['instrument_id'],
                'percentage__min': 0.0,
                'percentage__max': 0.0,
                'can_analyse__min': False,
                'can_analyse__max': False,
                'pending': list(pending_cont.all()),
                'not_validated': [],
                'pending_contributions': min(1.0, pending_cont.aggregate(
                    models.Avg('pending_contribution'))['pending_contribution__avg'] or 0.0),
                'not_validated_count': 0
            }
    for inst in missing:
        enrolments.append(missing[inst])
    return enrolments


@cache_memoize(30)
def get_missing_enrolment(learner_id, activity_id):
    """
        Get the missing enrolments for a learner and an activity
        :param learner_id: Learner id
        :param activity_id: Activity id
        :return: Object with missing instruments for this learner and activity
    """
    from .activity import Activity
    # Get related instances
    learner = Learner.objects.get(id=learner_id)
    activity = Activity.objects.get(id=activity_id)
    # Get the list of instruments and enrolment values
    enrolment_obj = {
        'missing_enrolments': False,
        'instruments': {},
        'activity_id': activity_id,
        'learner_id': learner_id
    }
    if activity.enabled:
        instrument_conf = activity.get_learner_instruments(learner)
        instruments = [inst.instrument.id for inst in instrument_conf]

        # Check enrolment status for this learner and activity
        missing_instruments = [inst.instrument.id for inst in instrument_conf if inst.instrument.requires_enrolment]
        for enrolment in learner.enrolment_status:
            if enrolment['instrument_id'] in instruments:
                missing_instruments.remove(enrolment['instrument_id'])
                enrolment_obj['instruments'][enrolment['instrument_id']] = enrolment
                if not enrolment['can_analyse__max']:
                    enrolment_obj['missing_enrolments'] = True
        if len(missing_instruments) > 0:
            enrolment_obj['missing_enrolments'] = True
            for inst_id in missing_instruments:
                enrolment_obj['instruments'][inst_id] = {
                    "instrument_id": inst_id,
                    "percentage__min": 0,
                    "percentage__max": 0,
                    "can_analyse__min": False,
                    "can_analyse__max": False,
                    "pending": [],
                    "pending_contributions": 0
                }
    return enrolment_obj


class Learner(InstitutionUser):
    """ Learner model. """
    learner_id = models.UUIDField(null=False, unique=True, blank=False,
                                  default=uuid.uuid4,
                                  help_text=_('Learner unique ID used to anonymize identity with external providers'))

    consent = models.ForeignKey(InformedConsent, null=True,
                                on_delete=models.SET_NULL, help_text=_('Current Informed consent of the learner'))

    consent_accepted = models.DateTimeField(null=True, blank=True,
                                            help_text=_('Date of acceptance of the current Informed Consent'))
    consent_rejected = models.DateTimeField(null=True, blank=True,
                                            help_text=_('Date of rejection of the current Informed Consent'))

    joined_at = models.DateTimeField(auto_now_add=True, help_text=_('Date the learner joined for this institution'))

    def __repr__(self):
        return "<Learner(learner_id='%r', consent_id='%r')>" % (
            self.learner_id, self.consent_id)

    @property
    def send(self):
        """
            SEND information for this learner
            :return: SEND summary
            :rtype: dict
        """
        return get_learner_send(self.id)

    @property
    def ic_status(self):
        """
            Informed Consent status for this learner
            :return: Informed Consent status
            :rtype: str
        """
        if self.institution.external_ic:
            return "VALID_EXTERNAL"
        if self.consent_rejected is not None:
            return "NOT_VALID_REJECTED"
        if self.consent is None:
            return "NOT_VALID_MISSING"
        return self.consent.status

    @property
    def enrolment_status(self):
        """
            Get the enrolment status for a learner
            :return: Aggregated values per instrument
        """
        return get_learner_enrolment(self.id)

    def missing_enrolments(self, activity_id):
        """
            Get the missing enrolments for a particular activity
            :param activity_id: Activity id
            :return: Aggregated values per instrument
        """
        return get_missing_enrolment(self.id, activity_id)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Authentication urls module """
from django.urls import path

from .views import app_view
from .views import logout_view
from .views import token_refresh_view
from .views import token_view
from .views import user_data_view
from .views import user_view

app_name = 'tesla_ce.apps.api'
urlpatterns = [
    path('login', user_view),
    path('logout', logout_view),
    path('profile', user_data_view),
    path('approle', app_view),
    path('token', token_view),
    path('token/refresh', token_refresh_view),
]



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Request model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .instrument import Instrument
from .message import Message
from .request import Request

RESULT_STATUS = (
    (0, _('Pending')),
    (1, _('Processed')),
    (2, _('Error')),
    (3, _('Timeout')),
    (4, _('Missing Provider')),
    (5, _('Missing Enrolment')),
    (6, _('Processing')),
    (7, _('Waiting External Service')),
)

RESULT_CODE = (
    (0, _('Pending')),
    (1, _('Ok')),
    (2, _('Warning')),
    (3, _('Alert'))
)


class RequestResult(BaseModel):
    """ Request result. """

    request = models.ForeignKey(Request, null=False,
                                on_delete=models.CASCADE, help_text=_('Request related to this result'))

    instrument = models.ForeignKey(Instrument, null=False,
                                   on_delete=models.CASCADE, help_text=_('Instrument related to this result'))

    status = models.IntegerField(choices=RESULT_STATUS, null=False,
                                 default=0,
                                 help_text=_('Status for this result'))

    result = models.FloatField(null=True, default=0.0,
                               help_text=_('Normalized result value, summarizing results from providers'))

    error_message = models.TextField(null=True,
                                     help_text=_("Error message when status is error"))

    message_code = models.ForeignKey(Message, null=True, default=None,
                                     on_delete=models.SET_NULL,
                                     help_text=_("Related message code"))

    code = models.IntegerField(choices=RESULT_CODE, null=False, default=0,
                               help_text=_("Result code provided after performing the verification process"))

    audit = models.FileField(max_length=250, null=False, blank=False,
                             help_text=_("Audit data for this request."))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __repr__(self):
        return "<RequestResult(request='%s', instrument='%s', status='%r', " \
               "result='%r', code='%s')>" % (
                   self.request.id, self.instrument.acronym, self.get_status_display(),
                   self.result, self.get_code_display()
        )



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA CE Activity use case test """
import pytest

from .utils.scenario import UseCaseScenario


@pytest.mark.django_db(transaction=False)
def test_activity_case_complete(rest_api_client, user_global_admin):
    """
        Complete example from institution creation to activity reports generation

        :param rest_api_client: Fixture that ensures that whole system is ready to use
        :param user_global_admin: Global administrator object
    """
    # Create an empty scenario with only the global user
    scenario = UseCaseScenario(
        global_admin = user_global_admin
    )
    assert scenario.get_global_admin() is not None
    assert scenario.check_global_admin()




#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Celery client module """
from __future__ import absolute_import

import os

from celery import Celery
from django.conf import settings
from kombu import Exchange
from kombu import Queue

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'tesla_ce.settings')
os.environ.setdefault('DJANGO_CONFIGURATION', 'Production')

import configurations
configurations.setup()

app = Celery('tesla_ce')

# Using a string here means the worker doesn't have to serialize
# the configuration object to child processes.
# - namespace='CELERY' means all celery-related configuration keys
#   should have a `CELERY_` prefix.
app.config_from_object('django.conf:settings', namespace='CELERY')

# Set queues to consume from
task_queues = tuple()
queues = []

if settings.TESLA_CONFIG.config.get('CELERY_QUEUES') is None or len(
        settings.TESLA_CONFIG.config.get('CELERY_QUEUES')
) == 0:
    # Set the default queue
    app.conf.task_default_queue = settings.TESLA_CONFIG.config.get('CELERY_QUEUE_DEFAULT')
elif len(settings.TESLA_CONFIG.config.get('CELERY_QUEUES')) == 1:
    if settings.TESLA_CONFIG.config.get('CELERY_QUEUES')[0] == '__all__':
        queues = [
            'ENROLMENT', 'ENROLMENT_STORAGE', 'ENROLMENT_VALIDATION', 'VERIFICATION',
            'ALERTS', 'REPORTING'
        ]
    else:
        # Set the default queue
        queue = settings.TESLA_CONFIG.config.get('CELERY_QUEUES')[0]
        app.conf.task_default_queue = settings.TESLA_CONFIG.config.get('CELERY_QUEUE_{}'.format(queue.upper()))
else:
    queues = settings.TESLA_CONFIG.config.get('CELERY_QUEUES')

if len(queues) > 0:
    queue_list = tuple()
    with app.broker_connection() as connection:
        channel = connection.default_channel
        for queue in queues:
            queue_name = settings.TESLA_CONFIG.config.get('CELERY_QUEUE_{}'.format(queue.upper()))
            new_queue = Queue(queue_name,
                              exchange=Exchange(queue_name, type='direct', connection=connection),
                              routing_key=queue_name)
            new_queue(channel).declare()
            queue_list += (new_queue, )
    app.conf.task_queues = queue_list

# Set routing
app.conf.task_routes = ([
    ('tesla_ce.tasks.requests.enrolment.create_sample', {
        'queue': settings.TESLA_CONFIG.config.get('CELERY_QUEUE_ENROLMENT_STORAGE')}
    ),
    ('tesla_ce.tasks.requests.enrolment.validate_request', {
        'queue': settings.TESLA_CONFIG.config.get('CELERY_QUEUE_ENROLMENT_VALIDATION')}
    ),
    ('tesla_ce.tasks.requests.enrolment.create_validation_summary', {
        'queue': settings.TESLA_CONFIG.config.get('CELERY_QUEUE_ENROLMENT_VALIDATION')}
    ),
    ('tesla_ce.tasks.requests.enrolment.enrol_learner', {
       'queue': settings.TESLA_CONFIG.config.get('CELERY_QUEUE_ENROLMENT')}
    ),
    ('tesla_ce.tasks.requests.verification.*', {
        'queue': settings.TESLA_CONFIG.config.get('CELERY_QUEUE_VERIFICATION')
    }),
    ('tesla_ce.tasks.requests.alerts.*', {'queue': settings.TESLA_CONFIG.config.get('CELERY_QUEUE_ALERTS')}),
    ('tesla_ce.tasks.reports.*', {'queue': settings.TESLA_CONFIG.config.get('CELERY_QUEUE_REPORTING')}),
],)

# Load task modules from all registered Django app configs.
app.autodiscover_tasks()

@app.task(bind=True)
def debug_task(self):
    print('Request: {0!r}'.format(self.request))



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Learner API Enrolment tasks module """
import simplejson
from celery import group
from celery.exceptions import MaxRetriesExceededError
from celery.exceptions import Reject
from django.core.files.base import ContentFile
from django.core.files.storage import default_storage

from tesla_ce import celery_app
from tesla_ce import models


@celery_app.task(ignore_result=True)
def create_sample(learner_id, path, instruments):
    try:
        # Store the sample
        learner = models.Learner.objects.get(learner_id=learner_id)

        # Check Informed Consent status
        if not learner.ic_status.startswith('VALID'):
            default_storage.delete(path)
            raise Reject("Missing informed consent", requeue=False)

        # Create the sample in the database
        sample = models.EnrolmentSample.objects.create(learner_id=learner.id,
                                                       data=path,
                                                       )
        # Add the instruments
        sample.instruments.set(instruments)
        if len(instruments) != sample.instruments.count():
            error = {
                "message": 'Invalid instruments',
            }
            default_storage.save('{}.error'.format(path), ContentFile(simplejson.dumps(error).encode('utf-8')))
            sample.status = 2
            sample.error_message = error["message"]
            sample.save()
            raise Reject("Invalid Instruments", requeue=False)

        # Once the sample has been stored, send validation task
        validate_request.apply_async((sample.learner.learner_id, sample.id, ))

        return sample.id
    except models.Learner.DoesNotExist:
        error = {
            "message": 'Invalid learner',
        }
        default_storage.save('{}.error'.format(path), ContentFile(simplejson.dumps(error).encode('utf-8')))
        raise Reject("Invalid Learner", requeue=False)


@celery_app.task(ignore_result=True)
def validate_request(learner_id, sample_id, validation_id=None):
    """
        Start validation of an enrolment sample
        :param learner_id: Learner identifier
        :type learner_id: str
        :param sample_id: Sample ID
        :type sample_id: int
        :param validation_id: Sample Validation ID. Used by provider workers
        :type validation_id: int
    """
    try:
        # Store the sample
        request = models.EnrolmentSample.objects.get(id=sample_id)

        # Get validators
        validators = []
        for instrument in request.instruments.all():
            inst_validators = instrument.provider_set.filter(enabled=True,
                                                             allow_validation=True,
                                                             validation_active=True).all()
            if len(inst_validators) == 0:
                validation = {
                    "status": "MISSING_PROVIDER",
                    "message": "There is no validation provider for instrument <{}>".format(instrument.name)
                }
                default_storage.save('{}.error'.format(request.data.name),
                                     ContentFile(simplejson.dumps(validation).encode('utf-8')))
                request.status = 2
                request.error_message = validation["message"]
                request.save()
                raise Reject("Missing Validation Providers", requeue=False)
            validators += list(inst_validators)

        # Perform the validation
        validation_group_tasks = []
        for validator in validators:
            valid_req = models.EnrolmentSampleValidation.objects.create(
                sample_id=sample_id,
                provider_id=validator.id
            )
            # TODO: Run validation task on each validator GROUP + RESULT TASK
            validation_group_tasks.append(validate_request.s(learner_id,
                                                             sample_id,
                                                             valid_req.id).set(queue=validator.queue))
        # Start tasks. Not using Chord to avoid requiring Providers to access the Database Backend
        group(validation_group_tasks).apply_async()
        #create_validation_summary.apply_async((learner_id, sample_id,), countdown=15)

    except models.EnrolmentSample.DoesNotExist:
        raise Reject("Invalid Sample", requeue=False)


@celery_app.task(bind=True, ignore_result=True, retry_backoff=True)
def create_validation_summary(self, learner_id, sample_id):
    """
        Summarize sample validation from the different validators
        :param learner_id: Learner identifier
        :type learner_id: str
        :param sample_id: Sample ID
        :type sample_id: int
    """

    # Get validations for this sample
    validations = models.EnrolmentSampleValidation.objects.filter(sample__learner__learner_id=learner_id,
                                                                  sample_id=sample_id)
    # TODO: Check if the current summary is newer than the last provider verification. (avoid updating finished summary)
    # Check if there are pending validations
    try:
        timeout = False
        if validations.filter(status=0).count() > 0:
            self.retry(countdown=15 + self.request.retries * 90, max_retries=5)
    except MaxRetriesExceededError:
        timeout = True

    # Get the sample
    sample = models.EnrolmentSample.objects.get(id=sample_id)

    # Intitialize validation data
    validation = {
        "status": None,
        "validations": {},
    }

    # Store the result
    valid = True
    for val_res in validations.all():
        if timeout and val_res.status == 0:
            val_res.status = 3
            val_res.save()
        if val_res.status > 1:
            valid = False
        info = None
        if val_res.info is not None and len(val_res.info.name) > 0:
            info = simplejson.loads(val_res.info.read().decode())
        validation['validations'][val_res.provider_id] = {
            'status': val_res.status,
            'status_code': val_res.get_status_display(),
            'error_message': val_res.error_message,
            'contribution': val_res.contribution,
            'info': info
        }

    if valid:
        validation['status'] = "VALID"
        sample.status = 1
        default_storage.save('{}.valid'.format(sample.data.name),
                             ContentFile(simplejson.dumps(validation).encode('utf-8')))
        # Start enrolment process
        enrol_learner.apply_async((learner_id, sample_id,))
    else:
        if timeout:
            validation['status'] = "TIMEOUT"
            sample.status = 3
            default_storage.save('{}.timeout'.format(sample.data.name),
                                 ContentFile(simplejson.dumps(validation).encode('utf-8')))
        else:
            validation['status'] = "ERROR"
            sample.status = 2
            default_storage.save('{}.error'.format(sample.data.name),
                                 ContentFile(simplejson.dumps(validation).encode('utf-8')))

    sample.save()


@celery_app.task(ignore_result=True)
def enrol_learner(learner_id, sample_id):
    """
        Perform learner enrolment for an instrument
        :param learner_id: Learner identifier
        :type learner_id: str
        :param sample_id: Sample ID
        :type sample_id: int
    """
    try:
        sample = models.EnrolmentSample.objects.get(id=sample_id)
        for instrument in sample.instruments.all():
            for provider in models.Provider.objects.filter(instrument_id=instrument.id, enabled=True).all():
                enrol_learner.apply_async((learner_id, sample_id,), queue=provider.queue)
    except models.EnrolmentSample.DoesNotExist:
        raise Reject(reason='Sample not found', requeue=False)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""VLE api serialize module."""
from rest_framework import exceptions
from rest_framework import serializers
from rest_framework import validators

from tesla_ce import get_default_client
from tesla_ce.apps.api.utils import JSONField
from tesla_ce.lib.exception import TeslaVaultException
from tesla_ce.models import VLE
from tesla_ce.models.vle import VLE_TYPE


class InstitutionVLESerializer(serializers.ModelSerializer):
    """VLE serialize model module."""
    _credentials = None

    type = serializers.ChoiceField(choices=VLE_TYPE)
    lti = JSONField(read_only=True)
    name = serializers.CharField(validators=[
                                    validators.UniqueValidator(queryset=VLE.objects.all())
                                 ])

    credentials = serializers.SerializerMethodField(default=None)

    class Meta:
        model = VLE
        exclude = ["created_at", "updated_at", "institution"]

    def validate(self, attrs):
        """
            Validate the given attributes
            :param attrs: Attributes parsed from request
            :type attrs: dict
            :return: Validated attributes
            :rtype: dict
        """
        # Add predefined fields
        attrs['institution_id'] = self.context['view'].kwargs['parent_lookup_institution_id']

        # Apply validators
        for validator in self.get_validators():
            validator(attrs, self)
        return super().validate(attrs)

    def create(self, validated_data):
        vle = super().create(validated_data)

        try:
            vle_info = get_default_client().register_vle(vle.get_type_display(), vle.name, vle.url,
                                                         vle.institution.acronym, vle.client_id, True)
        except TeslaVaultException:
            vle.delete()
            raise exceptions.PermissionDenied('Cannot register VLEs')

        self._credentials = vle_info

        return vle

    def get_credentials(self, object):
        return self._credentials



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""InformedConsent api serialize module."""
from rest_framework import serializers

from tesla_ce.models import InformedConsent
from .institution import InstitutionSerializer


class InstitutionInformedConsentVersionSerializer(serializers.ModelSerializer):
    """InformedConsent serialize model module."""
    class Meta:
        model = InformedConsent
        fields = "version"


class InstitutionInformedConsentSerializer(serializers.ModelSerializer):
    """InformedConsent serialize model module."""

    institution = InstitutionSerializer(read_only=True)
    institution_id = serializers.HiddenField(default=None, allow_null=True)

    class Meta:
        model = InformedConsent
        fields = "__all__"
        validators = [serializers.UniqueTogetherValidator(
            queryset=InformedConsent.objects.all(),
            fields=['institution_id', 'version']
        ), ]

    def validate(self, attrs):
        """
            Validate the given attributes
            :param attrs: Attributes parsed from request
            :type attrs: dict
            :return: Validated attributes
            :rtype: dict
        """
        # Add predefined fields
        attrs['institution_id'] = self.context['view'].kwargs['parent_lookup_institution_id']

        # Apply validators
        for validator in self.get_validators():
            validator(attrs, self)
        return super().validate(attrs)

    @classmethod
    def validate_version(self, value):
        """
            Validate the version format
            :param value: Provided version value
            :type value: str
            :return: Validated version
            :rtype: str
        """
        # Check version format
        try:
            version = [int(i) for i in value.split('.')]
            if len(version) != 3:
                raise ValueError()
        except ValueError:
            raise serializers.ValidationError('Invalid version. Expected X.Y.Z with X, Y and Z integer values')

        return '.'.join([str(i) for i in version])



#  Copyright (c) 2021 Mireia Bellot
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
import os
import logging
import pytest


# Check that ALL unauthenticated users are not allowed
def check_403(rest_api_client, str_path, users=None):
    """

    @param rest_api_client:
    @param str_path: path for checking non allowed users
    @param users: list of non allowed users
    """
    # Check that unauthenticated calls are not allowed
    if users is None:
        users = []
    response_403 = rest_api_client.get(str_path)
    assert response_403.status_code == 403

    # Check that global admins can list the instruments
    for usr in users:
        rest_api_client.force_authenticate(user=usr)
        response_403 = rest_api_client.get(str_path)
        assert response_403.status_code == 403


def print_log(args):
    """

    @param args: list of messages for logging. First item on list is the current testing module name.
    """
    module = ''
    if len(args) > 0:
        module = '[{}] '.format(args[0][0])
        # module = '[' + args[0][0] + '] '
    for i in range(1, len(args)):
        aux = module
        for j in args[i]:
            aux += str(j) + ' '
        logging.info(aux)


@pytest.mark.django_db
def get_rest_api_client(rest_api_client, str_path, module, message, status):
    """

    @param rest_api_client:
    @param str_path: GET path
    @param module: Testing module name
    @param message: Logging info message
    @param status: Testing status (example: 200, 404...)
    @return: response body
    """
    response = rest_api_client.get(str_path, format='json')
    body = response.json()
    str_log = [[module], ['Expected Status:', status],
               ['Received Status:', response.status_code], [message, body]]
    print_log(str_log)
    assert response.status_code == status
    return body


@pytest.mark.django_db
def post_rest_api_client(rest_api_client, str_path, str_data, module, message, status):
    """

    @param rest_api_client:
    @param str_path:
    @param str_data:
    @param module: Testing module name
    @param message: Logging info message
    @param status: Testing status (example: 200, 404...)
    @return: New item ID (-1 if creation failed)
    """
    response = rest_api_client.post(str_path, data=str_data, format='json')
    body = response.json()
    str_log = [[module], ['POST data:', str_data], ['Expected Status:', status],
               ['Received Status:', response.status_code], [message, body]]
    print_log(str_log)
    assert response.status_code == status
    if status == 201:
        new_item_id = response.json()['id']
        str_log.append(['NEW ITEM ID:', new_item_id])
    else:
        new_item_id = -1
    print_log(str_log)
    return new_item_id


@pytest.mark.django_db
def put_rest_api_client(rest_api_client, str_path, str_data, module, message, status):
    """

    @param rest_api_client:
    @param str_path:
    @param str_data:
    @param module: Testing module name
    @param message: Logging info message
    @param status: Testing status (example: 200, 404...)
    """
    response = rest_api_client.put(str_path, data=str_data, format='json')
    body = response.json()

    str_log = [[module], ['PUT item:', str_path], ['PUT data:', str_data],
               ['Expected Status:', status],
               ['Received Status:', response.status_code], [message, body]]
    print_log(str_log)
    assert response.status_code == status


@pytest.mark.django_db
def delete_rest_api_client(rest_api_client, str_path, module, message, status):
    """

    @param rest_api_client:
    @param str_path:
    @param module: Testing module name
    @param message: Logging info message
    @param status: Testing status (example: 200, 404...)
    """
    response = rest_api_client.delete(str_path)
    str_log = [[module], ['Expected Status:', status],
               ['Received Status:', response.status_code], [message, response],
               ['response.status_code', response.status_code]]
    print_log(str_log)
    assert response.status_code == status


def check_pagination(rest_api_client, body):
    body_aux = body
    offset = len(body_aux['results'])
    print_log([['Check pagination'], ['offset: ', offset]])
    while body_aux['next'] is not None:
        body_aux = get_rest_api_client(rest_api_client, body_aux['next'], 'Check pagination', 'Next page:', 200)
        offset = offset + len(body_aux['results'])
        print_log([['Check pagination'], ['offset: ', offset]])
    assert offset == body['count']


@pytest.mark.django_db
def getting_variables(body, institution_id) -> object:
    institution_empty = True
    id_non_existing_institution = 1000000
    n_institution = body['count']

    if n_institution > 0:
        institution_empty = False
        id_first_institution = body['results'][0]['id']
    else:
        id_first_institution = -1

    return {
            'institution_empty': institution_empty,
            'n_institution': n_institution,
            'id_first_institution': id_first_institution,
            'id_non_existing_institution': id_non_existing_institution
        }


@pytest.mark.django_db
def get_profile(rest_api_client, token):
    """
        Get authenticated user profile
        :param rest_api_client: API client
        :param token: Access token
        :return: Returned profile
    """
    rest_api_client.credentials(HTTP_AUTHORIZATION='JWT ' + token)
    profile_resp = rest_api_client.get('/api/v2/auth/profile')
    rest_api_client.credentials()

    return profile_resp


def get_provider_desc_file(provider):
    """
        Get the path to the file with the provider description
        :param provider: Provider name
        :return: Path to the file
    """
    return os.path.abspath(os.path.join(os.path.dirname(__file__),
                           '..', '..', 'providers', '{}.json'.format(provider)))


def get_module_auth_user(module):
    """
        Get a module authentication object from a object module instance
        :param module: Module object
        :return: Authentication object
    """
    from tesla_ce.models import VLE, Provider, AuthenticatedModule
    auth_object = None
    if isinstance(module, VLE) or isinstance(module, Provider):
        auth_object = AuthenticatedModule()
        auth_object.id = module.id
        auth_object.pk = module.id
        auth_object.module = module
        if isinstance(module, VLE):
            auth_object.type = 'vle'
        else:
            auth_object.type = 'provider'
    return auth_object



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Instrument views module """
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter

from tesla_ce.apps.api import permissions
from tesla_ce.apps.api.v2.serializers import InstrumentSerializer
from tesla_ce.models import Instrument


# pylint: disable=too-many-ancestors
class InstrumentViewSet(viewsets.ModelViewSet):
    """
    API endpoint that allows activity to be viewed or edited.
    """
    queryset = Instrument.objects.all().order_by('created_at')
    serializer_class = InstrumentSerializer
    permission_classes = [permissions.GlobalAdminPermission]
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    '''
    filterset_fields = ['activity_type', 'external_token', 'description', 'conf', 'vle']
    search_fields = ['activity_type', 'external_token', 'description', 'conf', 'vle']
    '''



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA CE scenario definition for Use Case tests """
from typing import List, Optional
from tesla_ce import models


class Activity:
    """
        Class defining a test activity
    """
    instruments = []


class Course:
    """
        Class defining a test course
    """
    # List of learners
    learners = []
    # List of instructors
    instructors = []


class Institution:
    """
        Class defining an institution
    """
    # Institution admin
    institution_admin: Optional[bool] = None
    # Institution SEND admin
    institution_send_admin: Optional[bool] = None
    # Institution Data admin
    institution_data_admin: Optional[bool] = None
    # Institution Legal admin
    institution_legal_admin: Optional[bool] = None
    # List of courses
    courses: List[Course] = []


class Context:
    """
        Define a test context
    """
    # Global admin
    global_admin: Optional[models.user.User] = None
    # Institution
    institutions: List[Institution] = []
    # Instruments

    def __init__(self, **kwargs):
        """
            Context constructor
        """
        pass


class UseCaseScenario:
    """
        Class defining a test scenario
    """
    # Current context
    current_context: Context = Context()
    # Expected context
    expected_context: Context = Context()

    def __init__(self, global_admin, **kwargs):
        """
            Scenario constructor
        """
        self.current_context.global_admin = global_admin
        self.expected_context.global_admin = global_admin

    def get_global_admin(self) -> Optional[models.user.User]:
        """
            Get the global admin user
            :return: Global admin user
        """
        return self.current_context.global_admin

    def check_global_admin(self):
        """
            Check that global admin is the expected
        """
        return self.current_context.global_admin.id == self.expected_context.global_admin.id



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" DJango command to perform the configuration of TeSLA system """
from django.core.management.base import CommandError

from tesla_ce.client import Client
from ..base import TeslaConfigCommand


class Command(TeslaConfigCommand):
    """ Reconfiguration command """
    help = 'Reconfigure TeSLA CE system'
    requires_system_checks = []

    def add_arguments(self, parser):
        """
            Define custom arguments for this command

            :param parser: Input command parser instance
        """
        parser.add_argument(
            '--local',
            action='store_true',
            help='Use the configuration file in current directory',
        )
        parser.add_argument(
            '--check',
            action='store_true',
            help='Check the configuration, but do not perform the installation',
        )

    def custom_handle(self):
        """
            Custom actions for this command
        """

        # Get configuration report
        report = self._get_configuration_report(self._options['check'])

        if report['valid']:
            # Performing reconfiguration
            self.stdout.write('Configuring TeSLA CE')

            # Initialize TeSLA
            self.client.initialize()



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Read Only Institution views module """
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter
from rest_framework.mixins import ListModelMixin
from rest_framework.mixins import UpdateModelMixin
from rest_framework_extensions.mixins import NestedViewSetMixin

from tesla_ce.apps.api import permissions
from tesla_ce.apps.api.v2.serializers import InstitutionSerializer
from tesla_ce.models import Institution


# pylint: disable=too-many-ancestors
class InstitutionViewSet(viewsets.ReadOnlyModelViewSet, NestedViewSetMixin, ListModelMixin, UpdateModelMixin):
    """
    API endpoint that allows access to institution data.
    """
    model = Institution
    permission_classes = [
        permissions.InstitutionAdminPermission |
        permissions.InstitutionMemberReadOnlyPermission |
        permissions.GlobalAdminReadOnlyPermission
    ]
    serializer_class = InstitutionSerializer
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    lookup_url_kwarg = 'institution_id'

    filterset_fields = ['acronym', 'name']
    search_fields = ['acronym', 'name']

    def get_queryset(self):
        qs = Institution.objects.all().order_by('created_at')
        if not self.request.user.is_staff:
            qs.filter(id=self.request.user.institution.id)
        return qs



# Generated by Django 3.0.3 on 2020-05-04 13:28

import uuid

import django.contrib.auth.models
import django.db.models.deletion
from django.conf import settings
from django.db import migrations
from django.db import models

import tesla_ce.models.enrolment
import tesla_ce.models.informed_consent_document


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('auth', '0011_update_proxy_permissions'),
    ]

    operations = [
        migrations.CreateModel(
            name='Activity',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('vle_activity_type', models.CharField(help_text='Activity type on the VLE.', max_length=250)),
                ('vle_activity_id', models.CharField(help_text='Activity id on the VLE.', max_length=250)),
                ('description', models.TextField(blank=True, help_text='Activity description.', null=True)),
                ('name', models.CharField(blank=True, help_text='Activity name.', max_length=255, null=True)),
                ('enabled', models.BooleanField(blank=None, default=False, help_text='Whether this activity is enabled or not', null=None)),
                ('start', models.DateTimeField(blank=True, help_text='When activity starts', null=True)),
                ('end', models.DateTimeField(blank=True, help_text='When activity ends', null=True)),
                ('conf', models.TextField(blank=True, help_text='Activity conf.', null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name='AssessmentSession',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('connector', models.FileField(help_text='Connector JS file for this session', upload_to='')),
                ('started_at', models.DateTimeField(auto_now_add=True, help_text='Date when this assessment session started')),
                ('checked_at', models.DateTimeField(blank=True, help_text='Last time this assessment session has been checked', null=True)),
                ('closed_at', models.DateTimeField(blank=True, help_text='Date the assessment session has been closed', null=True)),
                ('activity', models.ForeignKey(help_text='Activity of this session', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Activity')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Course',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(help_text='Course code.', max_length=250)),
                ('vle_course_id', models.CharField(help_text='Course id on the VLE.', max_length=250)),
                ('description', models.TextField(blank=True, help_text='Course description.', null=True)),
                ('start', models.DateTimeField(blank=True, help_text='When course starts', null=True)),
                ('end', models.DateTimeField(blank=True, help_text='When course ends', null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
            ],
        ),
        migrations.CreateModel(
            name='Institution',
            fields=[
                ('id', models.AutoField(help_text='Institution Id', primary_key=True, serialize=False, unique=True)),
                ('acronym', models.CharField(help_text='Institution acronym', max_length=255, unique=True)),
                ('name', models.TextField(blank=None, help_text='Name of the institution', null=None)),
                ('external_ic', models.BooleanField(default=False, help_text='Informed Consent is managed externally to TeSLA', null=None)),
                ('mail_domain', models.CharField(blank=True, default=None, help_text='Accepted mail domains for this institution', max_length=255, null=True)),
                ('disable_vle_learner_creation', models.BooleanField(default=False, help_text='If enabled, VLEs cannot create learners', null=None)),
                ('disable_vle_instructor_creation', models.BooleanField(default=False, help_text='If enabled, VLE cannot create instructors', null=None)),
                ('disable_vle_user_creation', models.BooleanField(default=False, help_text='If enabled, VLE cannot create institution users', null=None)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='InstitutionUser',
            fields=[
                ('user_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to=settings.AUTH_USER_MODEL)),
                ('uid', models.CharField(help_text='Unique User Identifier for the institution', max_length=255)),
                ('login_allowed', models.BooleanField(default=False, help_text='Whether this user can login with user/password')),
                ('institution', models.ForeignKey(help_text='Institution of the user', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Institution')),
            ],
            options={
                'unique_together': {('institution', 'uid')},
            },
            bases=('auth.user',),
            managers=[
                ('objects', django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name='Instrument',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Instrument name.', max_length=250)),
                ('acronym', models.CharField(help_text='Instrument Acronym.', max_length=30, unique=True)),
                ('queue', models.TextField(help_text='Queue this instrument listens to.', unique=True)),
                ('enabled', models.BooleanField(default=False, help_text='The instrument is enabled')),
                ('requires_enrolment', models.BooleanField(default=False, help_text='Whether this instrument requires enrolment')),
                ('description', models.TextField(blank=True, default=None, help_text='Description of the instrument.', null=True)),
                ('identity', models.BooleanField(default=False, help_text='This instrument contributes to the learner identity verification')),
                ('originality', models.BooleanField(default=False, help_text='This instrument contributes to the assessment originality verification')),
                ('authorship', models.BooleanField(default=False, help_text='This instrument contributes to the assessment authorship verification')),
                ('integrity', models.BooleanField(default=False, help_text='This instrument contributes to the assessment integrity verification')),
                ('options_schema', models.TextField(default=None, help_text='Schema for instrument options', null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Message',
            fields=[
                ('code', models.CharField(help_text='Message code', max_length=50, primary_key=True, serialize=False)),
                ('datatype', models.CharField(default='html', help_text='Type of the message', max_length=50)),
                ('description', models.TextField(default=None, help_text='Description of the message for translation', null=True)),
                ('meaning', models.TextField(default=None, help_text='Meaning of the message for translation', null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Provider',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Provider name.', max_length=250)),
                ('description', models.TextField(blank=True, default=None, help_text='Provider description.', null=True)),
                ('url', models.CharField(default=None, help_text='Provider url.', max_length=250, null=True)),
                ('version', models.CharField(help_text='Provider version.', max_length=15)),
                ('acronym', models.CharField(help_text='Provider Acronym.', max_length=30, unique=True)),
                ('queue', models.CharField(help_text='Queue where provider listens for requests.', max_length=50, unique=True)),
                ('enabled', models.BooleanField(default=False, help_text='Whether this provider is enabled to be used')),
                ('allow_validation', models.BooleanField(default=False, help_text='Whether this provider provides validation feature for data')),
                ('validation_active', models.BooleanField(default=False, help_text='Whether this provider will be used for validation')),
                ('alert_below', models.DecimalField(decimal_places=2, default=None, help_text='Mark as alert for result values below this threshold.', max_digits=4, null=True)),
                ('warning_below', models.DecimalField(decimal_places=2, default=None, help_text='Mark as warning for result values below this threshold.', max_digits=4, null=True)),
                ('inverted_polarity', models.BooleanField(default=False, help_text='If enabled, good values are lower values')),
                ('image', models.CharField(help_text='Provider Docker image.', max_length=250)),
                ('has_service', models.BooleanField(default=False, help_text='Whether this provider starts a service and must be balanced')),
                ('service_port', models.IntegerField(default=None, help_text='Port where service is listening', null=True)),
                ('options_schema', models.TextField(blank=True, default=None, help_text='Schema for provider options.', null=True)),
                ('options', models.TextField(blank=True, default=None, help_text='Provider options.', null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('instrument', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, to='tesla_ce.Instrument')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ReportActivity',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('identity_level', models.SmallIntegerField(choices=[(0, 'Pending'), (1, 'No Information'), (2, 'Ok'), (3, 'Warning'), (4, 'Alert')], default=0, help_text='Alert level for learner identity.')),
                ('content_level', models.SmallIntegerField(choices=[(0, 'Pending'), (1, 'No Information'), (2, 'Ok'), (3, 'Warning'), (4, 'Alert')], default=0, help_text='Alert level for content authorship.')),
                ('integrity_level', models.SmallIntegerField(choices=[(0, 'Pending'), (1, 'No Information'), (2, 'Ok'), (3, 'Warning'), (4, 'Alert')], default=0, help_text='Alert level for system integrity.')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('activity', models.ForeignKey(help_text='Activity related to this report.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Activity')),
            ],
        ),
        migrations.CreateModel(
            name='Request',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.IntegerField(choices=[(0, 'Stored'), (1, 'Scheduled'), (2, 'Processing'), (3, 'Processed'), (4, 'Error'), (5, 'Timeout'), (6, 'Missing Provider')], default=0, help_text='Status for this request')),
                ('data', models.FileField(help_text='Data path on storage.', max_length=250, upload_to='')),
                ('error_message', models.TextField(help_text='Error message when status is error', null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('activity', models.ForeignKey(default=None, null=True, on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Activity')),
                ('instruments', models.ManyToManyField(to='tesla_ce.Instrument')),
                ('message_code', models.ForeignKey(default=None, help_text='Related message code', null=True, on_delete=django.db.models.deletion.SET_NULL, to='tesla_ce.Message')),
                ('session', models.ForeignKey(default=None, help_text='Assessment session for this request', null=True, on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.AssessmentSession')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SENDCategory',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.TextField(help_text='Category description.')),
                ('data', models.TextField(help_text='Category configuration.')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('institution', models.ForeignKey(help_text='Institution this category is defined for', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Institution')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Instructor',
            fields=[
                ('institutionuser_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='tesla_ce.InstitutionUser')),
            ],
            options={
                'verbose_name': 'user',
                'verbose_name_plural': 'users',
                'abstract': False,
            },
            bases=('tesla_ce.institutionuser',),
            managers=[
                ('objects', django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name='Learner',
            fields=[
                ('institutionuser_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='tesla_ce.InstitutionUser')),
                ('learner_id', models.UUIDField(default=uuid.uuid4, help_text='Learner unique ID used to anonymize identity with external providers', unique=True)),
                ('consent_accepted', models.DateTimeField(blank=True, help_text='Date of acceptance of the current Informed Consent', null=True)),
                ('consent_rejected', models.DateTimeField(blank=True, help_text='Date of rejection of the current Informed Consent', null=True)),
                ('joined_at', models.DateTimeField(auto_now_add=True, help_text='Date the learner joined for this institution')),
            ],
            options={
                'verbose_name': 'user',
                'verbose_name_plural': 'users',
                'abstract': False,
            },
            bases=('tesla_ce.institutionuser',),
            managers=[
                ('objects', django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name='VLE',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='VLE unique name.', max_length=250, unique=True)),
                ('type', models.IntegerField(choices=[(0, 'MOODLE')], default=0, help_text='Type of the VLE.')),
                ('url', models.TextField(blank=True, help_text='VLE url.', null=True)),
                ('lti', models.TextField(blank=True, help_text='VLE lti configuration.', null=True)),
                ('client_id', models.CharField(default=None, help_text='LTI 1.3 Client ID.', max_length=250, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('institution', models.ForeignKey(help_text='Institution this VLE belongs to.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Institution')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RequestResult',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.IntegerField(choices=[(0, 'Pending'), (1, 'Processed'), (2, 'Error'), (3, 'Timeout'), (4, 'Missing Provider'), (5, 'Missing Enrolment')], default=0, help_text='Status for this result')),
                ('result', models.DecimalField(decimal_places=2, help_text='Normalized result value, summarizing results from providers', max_digits=3, null=True)),
                ('error_message', models.TextField(help_text='Error message when status is error', null=True)),
                ('code', models.IntegerField(choices=[(0, 'Pending'), (1, 'Ok'), (2, 'Warning'), (3, 'Alert')], default=0, help_text='Result code provided after performing the verification process')),
                ('audit', models.FileField(help_text='Audit data for this request.', max_length=250, upload_to='')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('instrument', models.ForeignKey(help_text='Instrument related to this result', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Instrument')),
                ('message_code', models.ForeignKey(default=None, help_text='Related message code', null=True, on_delete=django.db.models.deletion.SET_NULL, to='tesla_ce.Message')),
                ('request', models.ForeignKey(help_text='Request related to this result', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Request')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='RequestProviderResult',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.IntegerField(choices=[(0, 'Pending'), (1, 'Processed'), (2, 'Error'), (3, 'Timeout'), (4, 'Missing Provider'), (5, 'Missing Enrolment')], default=0, help_text='Status for this result')),
                ('result', models.DecimalField(decimal_places=2, help_text='Normalized result value', max_digits=3, null=True)),
                ('error_message', models.TextField(help_text='Error message when status is error', null=True)),
                ('code', models.IntegerField(choices=[(0, 'Pending'), (1, 'Ok'), (2, 'Warning'), (3, 'Alert')], default=0, help_text='Result code provided after performing the verification process')),
                ('audit', models.FileField(help_text='Audit data for this result.', max_length=250, upload_to='')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('message_code', models.ForeignKey(default=None, help_text='Related message code', null=True, on_delete=django.db.models.deletion.SET_NULL, to='tesla_ce.Message')),
                ('provider', models.ForeignKey(help_text='Provider related to this result', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Provider')),
                ('request', models.ForeignKey(help_text='Request related to this result', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Request')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='InformedConsent',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('version', models.CharField(help_text='Informed consent version.', max_length=250)),
                ('valid_from', models.DateTimeField(help_text='Informed consent valid from')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('institution', models.ForeignKey(help_text='Institution of the informed consent', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Institution')),
            ],
            options={
                'unique_together': {('institution', 'version')},
            },
        ),
        migrations.CreateModel(
            name='EnrolmentSample',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.IntegerField(choices=[(0, 'Stored'), (1, 'Valid'), (2, 'Error'), (3, 'Timeout'), (4, 'Missing Validator')], default=0, help_text='Status for this sample')),
                ('error_message', models.TextField(blank=None, default=None, help_text='Error message when status is error', null=True)),
                ('data', models.FileField(help_text='Data path on storage.', max_length=250, upload_to='')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('instruments', models.ManyToManyField(help_text='Instruments this sample is collected for', to='tesla_ce.Instrument')),
                ('message_code', models.ForeignKey(default=None, help_text='Related message code', null=True, on_delete=django.db.models.deletion.SET_NULL, to='tesla_ce.Message')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='CourseGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='Group name.', max_length=250)),
                ('description', models.TextField(blank=True, help_text='Group description.', null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('courses', models.ManyToManyField(blank=True, help_text='Courses in this group.', to='tesla_ce.Course')),
                ('institution', models.ForeignKey(help_text='Institution for this group.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Institution')),
                ('parent', models.ForeignKey(blank=True, help_text='Parent group.', null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='tesla_ce.CourseGroup')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='course',
            name='vle',
            field=models.ForeignKey(help_text='Course VLE.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.VLE'),
        ),
        migrations.CreateModel(
            name='ActivityInstrument',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('required', models.BooleanField(help_text='Is instrument required?')),
                ('active', models.BooleanField(help_text='Is instrument active?')),
                ('options', models.TextField(blank=True, help_text='Instrument options?', null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('activity', models.ForeignKey(help_text='Activity.', on_delete=django.db.models.deletion.CASCADE, related_name='configuration', to='tesla_ce.Activity')),
                ('alternative_to', models.ForeignKey(help_text='Primary instrument to be used.', null=True, on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.ActivityInstrument')),
                ('instrument', models.ForeignKey(help_text='Activity.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Instrument')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='activity',
            name='course',
            field=models.ForeignKey(help_text='Activity VLE.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Course'),
        ),
        migrations.AddField(
            model_name='activity',
            name='vle',
            field=models.ForeignKey(help_text='Activity VLE.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.VLE'),
        ),
        migrations.AddField(
            model_name='request',
            name='learner',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Learner'),
        ),
        migrations.CreateModel(
            name='ReportActivityInstrument',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('enrolment', models.SmallIntegerField(help_text='Enrolment percentage', null=None)),
                ('confidence', models.SmallIntegerField(help_text='Confidence percentage', null=None)),
                ('result', models.SmallIntegerField(help_text='Result percentage', null=None)),
                ('identity_level', models.SmallIntegerField(choices=[(0, 'Pending'), (1, 'No Information'), (2, 'Ok'), (3, 'Warning'), (4, 'Alert')], default=0, help_text='Alert level for learner identity.')),
                ('content_level', models.SmallIntegerField(choices=[(0, 'Pending'), (1, 'No Information'), (2, 'Ok'), (3, 'Warning'), (4, 'Alert')], default=0, help_text='Alert level for content authorship.')),
                ('integrity_level', models.SmallIntegerField(choices=[(0, 'Pending'), (1, 'No Information'), (2, 'Ok'), (3, 'Warning'), (4, 'Alert')], default=0, help_text='Alert level for system integrity.')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('instrument', models.ForeignKey(help_text='Instrument related to this report detail.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Instrument')),
                ('report', models.ForeignKey(help_text='Related Activity Report.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.ReportActivity')),
            ],
            options={
                'unique_together': {('report', 'instrument')},
            },
        ),
        migrations.AddField(
            model_name='reportactivity',
            name='learner',
            field=models.ForeignKey(help_text='Learner related to this report.', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Learner'),
        ),
        migrations.CreateModel(
            name='MessageLocale',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('locale', models.CharField(help_text='Locale', max_length=10)),
                ('message', models.TextField(help_text='Message')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('code', models.ForeignKey(help_text='Message code', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Message')),
            ],
            options={
                'unique_together': {('code', 'locale')},
            },
        ),
        migrations.AddField(
            model_name='learner',
            name='consent',
            field=models.ForeignKey(help_text='Current Informed consent of the learner', null=True, on_delete=django.db.models.deletion.SET_NULL, to='tesla_ce.InformedConsent'),
        ),
        migrations.CreateModel(
            name='InformedConsentDocument',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('language', models.CharField(max_length=30)),
                ('html', models.TextField(blank=True, help_text='HTML version of IC', null=True)),
                ('pdf', models.FileField(blank=True, help_text='PDF version of IC.', null=True, upload_to=tesla_ce.models.informed_consent_document.get_upload_path)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('consent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.InformedConsent')),
            ],
            options={
                'unique_together': {('consent', 'language')},
            },
        ),
        migrations.CreateModel(
            name='EnrolmentSampleValidation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('status', models.IntegerField(choices=[(0, 'Validating'), (1, 'Valid'), (2, 'Error'), (3, 'Timeout')], default=0, help_text='Validation status for this sample')),
                ('info', models.FileField(help_text='Validation information path on storage.', max_length=250, upload_to='')),
                ('contribution', models.DecimalField(decimal_places=2, default=None, help_text='Estimated contribution of this sample to the enrolment model', max_digits=5, null=True)),
                ('error_message', models.TextField(blank=None, default=None, help_text='Error message when status is error', null=True)),
                ('included', models.BooleanField(default=False, help_text='Whether this sample is included in current model')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('message_code', models.ForeignKey(default=None, help_text='Related message code', null=True, on_delete=django.db.models.deletion.SET_NULL, to='tesla_ce.Message')),
                ('provider', models.ForeignKey(help_text='Provider that performed the validation of the sample', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Provider')),
                ('sample', models.ForeignKey(help_text='Related Sample', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.EnrolmentSample')),
            ],
            options={
                'unique_together': {('sample', 'provider')},
            },
        ),
        migrations.AddField(
            model_name='enrolmentsample',
            name='learner',
            field=models.ForeignKey(help_text='Related learner', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Learner'),
        ),
        migrations.AddField(
            model_name='course',
            name='instructors',
            field=models.ManyToManyField(blank=True, help_text='Course instructors.', to='tesla_ce.Instructor'),
        ),
        migrations.AddField(
            model_name='course',
            name='learners',
            field=models.ManyToManyField(blank=True, help_text='Course learners.', to='tesla_ce.Learner'),
        ),
        migrations.AddField(
            model_name='assessmentsession',
            name='learner',
            field=models.ForeignKey(help_text='Learner of this session', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Learner'),
        ),
        migrations.CreateModel(
            name='Alert',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('level', models.IntegerField(choices=[(0, 'Info'), (1, 'Warning'), (2, 'Alert'), (3, 'Error')], default=0, help_text='Level of the alert')),
                ('data', models.FileField(help_text='Data path on storage.', max_length=250, upload_to='')),
                ('raised_at', models.DateTimeField(help_text='When the alert was raised.')),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('activity', models.ForeignKey(help_text='Related activity', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Activity')),
                ('instruments', models.ManyToManyField(help_text='Instruments related to this alert', to='tesla_ce.Instrument')),
                ('message_code', models.ForeignKey(default=None, help_text='Related message code', null=True, on_delete=django.db.models.deletion.SET_NULL, to='tesla_ce.Message')),
                ('session', models.ForeignKey(default=None, help_text='Related assessment session', null=True, on_delete=django.db.models.deletion.SET_NULL, to='tesla_ce.AssessmentSession')),
                ('learner', models.ForeignKey(help_text='Related learner', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Learner')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AlterUniqueTogether(
            name='activity',
            unique_together={('vle', 'vle_activity_type', 'vle_activity_id')},
        ),
        migrations.CreateModel(
            name='SENDLearner',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('expires_at', models.DateTimeField(null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.SENDCategory')),
                ('learner', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Learner')),
            ],
            options={
                'unique_together': {('learner', 'category')},
            },
        ),
        migrations.AlterUniqueTogether(
            name='reportactivity',
            unique_together={('activity', 'learner')},
        ),
        migrations.CreateModel(
            name='Enrolment',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('percentage', models.DecimalField(decimal_places=2, help_text='Percentage of enrolment for this learner', max_digits=5)),
                ('can_analyse', models.BooleanField(default=False, help_text='This provider can verify learner samples')),
                ('locked_at', models.DateTimeField(default=None, help_text='This model is locked for an update process', null=True)),
                ('locked_by', models.UUIDField(blank=None, default=None, help_text='Task blocking this model', null=True)),
                ('model', models.FileField(default=None, help_text='Path to stored model', max_length=250, null=True, upload_to=tesla_ce.models.enrolment.get_upload_path)),
                ('model_samples', models.IntegerField(default=0, help_text='Number of enrolment samples used for the model', null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(auto_now=True)),
                ('provider', models.ForeignKey(help_text='Related provider', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Provider')),
                ('learner', models.ForeignKey(help_text='Related learner', on_delete=django.db.models.deletion.CASCADE, to='tesla_ce.Learner')),
            ],
            options={
                'unique_together': {('learner', 'provider')},
            },
        ),
        migrations.AlterUniqueTogether(
            name='course',
            unique_together={('vle', 'vle_course_id')},
        ),
    ]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Institution model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel


class Institution(BaseModel):
    """ Institution model. """
    id = models.AutoField(null=False, unique=True, blank=False, primary_key=True,
                          help_text=_('Institution Id'))
    acronym = models.CharField(max_length=255, null=False, blank=False, unique=True,
                               help_text=_('Institution acronym'))
    name = models.TextField(blank=None, null=None, help_text=_('Name of the institution'))
    external_ic = models.BooleanField(null=None, default=False,
                                      help_text=_('Informed Consent is managed externally to TeSLA'))
    mail_domain = models.CharField(max_length=255, null=True, blank=True, default=None,
                                   help_text=_('Accepted mail domains for this institution'))
    disable_vle_learner_creation = models.BooleanField(null=None, default=False,
                                                       help_text=_('If enabled, VLEs cannot create learners'))
    disable_vle_instructor_creation = models.BooleanField(null=None, default=False,
                                                          help_text=_('If enabled, VLE cannot create instructors'))
    disable_vle_user_creation = models.BooleanField(null=None, default=False,
                                                    help_text=_('If enabled, VLE cannot create institution users'))
    allow_learner_report = models.BooleanField(null=None, default=False,
                                               help_text=_('Learners can access their reports'))
    allow_learner_audit = models.BooleanField(null=None, default=False,
                                              help_text=_('Learners can access the audit data of their reports'))
    allow_valid_audit = models.BooleanField(null=None, default=False,
                                            help_text=_('Audit data is available even when results are valid'))

    allowed_domains = models.TextField(null=True, blank=True, default=None,
                                       help_text=_('Allowed domains for redirection'))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __repr__(self):
        return "<Institution(id='%r', acronym='%r', name='%r', external_ic='%r')>" % (
            self.id, self.acronym, self.name, self.external_ic)



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Test package for VLE API """
import pytest
from tests.utils import get_module_auth_user

@pytest.mark.django_db
def test_api_vle(rest_api_client, institution_course_test_case):
    # Ensure VLE list is not allowed
    vle_list_no_auth_response = rest_api_client.get('/api/v2/vle/')
    assert vle_list_no_auth_response.status_code == 403

    # Authenticate with VLE object
    rest_api_client.force_authenticate(user=get_module_auth_user(institution_course_test_case['vle']))

    # Check that is not possible to list the VLEs
    vle_list_auth_response = rest_api_client.get('/api/v2/vle/')
    assert vle_list_auth_response.status_code == 403

    # Get VLE information
    vle_id = institution_course_test_case['vle'].id
    vle_info_auth_response = rest_api_client.get('/api/v2/vle/{}/'.format(vle_id))
    assert vle_info_auth_response.status_code == 200
    assert vle_info_auth_response.data['id'] == vle_id

#TODO VLE
#TODO Read VLE information
#TODO List Courses in a VLE
#TODO Create a new Course
#TODO Read Course information
#TODO Update Course information
#TODO Delete a Course
#TODO List Activities in a VLE Course
#TODO Create a new Activity
#TODO Read Activity information
#TODO Update Activity information
#TODO Delete an Activity
#TODO List Course learners --> TO BE DONE!
#TODO Add a learner to a course --> TO BE DONE!
#TODO Remove a learner from a course --> TO BE DONE!
#TODO List Course instructors --> TO BE DONE!
#TODO Add an instructor to a course --> TO BE DONE!
#TODO Remove an instructor from a course --> TO BE DONE!
#TODO TODO: Add assessment documentation --> TO BE DONE!



